"""Coppied from https://github.com/predict-idlab/landmark-uq"""

from typing import Optional

import cv2
import numpy as np
import torch
from scipy.stats import norm, spearmanr  # type: ignore
from scipy.stats.mstats import mquantiles  # type: ignore
from skimage import measure  # type: ignore
from tqdm import tqdm  # type: ignore

from landmarker.utils import pixel_to_unit_numpy


def mahalanobis_distance(pred, pred_cov, target):
    """Calculate the Mahalanobis distance between the predicted mean and the target.

    Args:
        pred: numpy array of shape (..., D) containing predicted means
        pred_cov: numpy array of shape (..., D, D) containing covariance matrices
        target: numpy array of shape (..., D) containing target values

    Returns:
        numpy array of shape (...) containing Mahalanobis distances
    """
    # Add a dimension to make diff shape (..., 1, D)
    diff = np.expand_dims(pred - target, axis=-2)
    # Calculate inverse of covariance matrix
    cov_inv = np.linalg.inv(pred_cov)
    # Calculate Mahalanobis distance
    # diff @ cov_inv @ diff.T -> shape (..., 1, 1)
    dist = np.sqrt(
        (diff @ cov_inv @ np.transpose(diff, axes=(*range(diff.ndim - 2), -1, -2)))
        .squeeze(axis=-1)
        .squeeze(axis=-1)
    )
    return dist


def euclidean_distance(pred, target):
    """Calculate the Euclidean distance between the predicted mean and the target.

    Args:
        pred: numpy array of shape (..., D)
        target: numpy array of shape (..., D)

    Returns:
        numpy array of shape (...)
    """
    return np.sqrt(np.sum((pred - target) ** 2, axis=-1))


def absolute_error_max(pred, target, sigmas=None):
    """Calculate the absolute error between the predicted mean and the target. The
    maximum absolute error is taken over the last dimension.

    Args:
        pred: numpy array of any shape
        target: numpy array of the same shape as pred
        sigmas: numpy array of the same shape as pred, containing the standard deviation

    Returns:
        numpy array of the same shape as inputs
    """
    if sigmas is None:
        return np.abs(pred - target).max(axis=-1)
    return (np.abs(pred - target) / sigmas).max(axis=-1)


class ConformalRegressorMahalanobis:
    def __init__(self, spatial_dims=2, nb_landmarks=1):
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks
        self.alphas = None

    def check_input(self, pred, pred_cov):
        if pred.ndim != 3:
            raise ValueError("pred must have shape (N, C, D)")
        if pred_cov.ndim != 4:
            raise ValueError("pred_cov must have shape (N, C, D, D)")
        if pred.shape[2] != pred_cov.shape[2] and self.spatial_dims != pred.shape[2]:
            raise ValueError("pred and pred_cov must have the same second dimension")
        if pred_cov.shape[2] != pred_cov.shape[3]:
            raise ValueError("pred_cov must be square")
        if pred.shape[0] != pred_cov.shape[0]:
            raise ValueError("pred and pred_cov must have the same first dimension")
        if pred.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        if pred_cov.shape[1] != self.nb_landmarks:
            raise ValueError("pred_cov must have the same second dimension as nb_landmarks")

    def fit(self, pred, pred_cov, target):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(
            pred, pred_cov
        )  # check if the target has the same shape as the calibration set
        assert pred.shape == target.shape
        self.alphas = mahalanobis_distance(pred, pred_cov, target)
        # Sort the non-conformity scores in descending order
        self.alphas = np.sort(self.alphas, axis=0)[::-1]

    def bound(self, pred_cov, confidence=0.95):
        """Get the bound for the given confidence level."""
        if self.alphas is None:
            raise ValueError("You must call fit before calling predict")
        return pred_cov * (
            self.alphas[int((1 - confidence) * (len(self.alphas) + 1)) - 1].reshape((1, -1, 1, 1))
            ** 2
        )

    def area_prediction_region(self, pred_cov, spacing=None, confidence=0.95):
        """
        Calculate the area of the prediction region, accounting for pixel spacing.

        Parameters:
        pred_cov (ndarray): The predicted covariance matrix.
        spacing (ndarray, optional): The spacing between pixels in millimeters,
                         with shape (Batch_size, spatial_size).
                         Defaults to None.
        confidence (float, optional): The confidence level for the prediction region.
                         Defaults to 0.95.

        Returns:
        float: The area of the prediction region.
        """
        pred_cov_scaled = self.bound(pred_cov, confidence)
        if spacing is not None:
            if isinstance(spacing, torch.Tensor):
                spacing = spacing.detach().numpy()
            assert spacing.shape == (pred_cov.shape[0], self.spatial_dims)
            # Make a diagonal matrix with the pixel spacing (Batch_size, spatial_size, spatial_size)
            # e.g. for 2D: [[s1, 0], [0, s2]], for 3D: [[s1, 0, 0], [0, s2, 0], [0, 0, s3]]
            spacing_matrix = np.stack(
                [np.diag(spacing[i]) for i in range(spacing.shape[0])], axis=0
            ).reshape((pred_cov.shape[0], 1, self.spatial_dims, self.spatial_dims))
            # Scale the covariance matrix by the pixel spacing
            pred_cov_scaled = (
                spacing_matrix @ pred_cov_scaled @ np.transpose(spacing_matrix, (0, 1, 3, 2))
            )
        eig = np.linalg.eigvalsh(pred_cov_scaled)
        return np.pi * np.prod(np.sqrt(eig), axis=-1)

    def predict(self, pred, pred_cov, confidence=0.95):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        self.check_input(pred, pred_cov)
        return pred, self.bound(pred_cov, confidence)

    def predict_contour(self, pred, pred_cov, confidence=0.95, sample_points=200):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        self.check_input(pred, pred_cov)

        if self.spatial_dims == 2:
            angles = np.linspace(0, 2 * np.pi, sample_points)
            directions = np.stack((np.cos(angles), np.sin(angles)), axis=-1)
            directions = directions.reshape((1, -1, 1, 1, self.spatial_dims))
            error_bound = directions @ np.linalg.cholesky(self.bound(pred_cov, confidence)).reshape(
                (-1, 1, self.nb_landmarks, self.spatial_dims, self.spatial_dims)
            )
            return pred.reshape(
                (-1, 1, self.nb_landmarks, self.spatial_dims)
            ) + error_bound.reshape((-1, directions.shape[1], self.nb_landmarks, self.spatial_dims))
        elif self.spatial_dims == 3:
            eigenvalues, eigenvectors = np.linalg.eigh(self.bound(pred_cov, confidence))

            # Create a grid of points for a unit sphere
            u = np.linspace(0, 2 * np.pi, sample_points)
            v = np.linspace(0, np.pi, sample_points // 2)
            x = np.outer(np.cos(u), np.sin(v))
            y = np.outer(np.sin(u), np.sin(v))
            z = np.outer(np.ones_like(u), np.cos(v))

            # Scale and rotate the unit sphere to match the ellipsoid
            radii = np.sqrt(eigenvalues)
            transform = eigenvectors @ np.stack(
                [
                    np.stack(
                        [radii[..., 0], np.zeros_like(radii[..., 0]), np.zeros_like(radii[..., 0])],
                        axis=-1,
                    ),
                    np.stack(
                        [np.zeros_like(radii[..., 0]), radii[..., 1], np.zeros_like(radii[..., 0])],
                        axis=-1,
                    ),
                    np.stack(
                        [np.zeros_like(radii[..., 0]), np.zeros_like(radii[..., 0]), radii[..., 2]],
                        axis=-1,
                    ),
                ],
                axis=-2,
            )
            ellipsoid = transform @ np.array([z.ravel(), y.ravel(), x.ravel()])
            # Reshape to 3D grid
            x_ellipsoid = ellipsoid[..., 2, :].reshape((-1, self.nb_landmarks, *x.shape))
            y_ellipsoid = ellipsoid[..., 1, :].reshape((-1, self.nb_landmarks, *y.shape))
            z_ellipsoid = ellipsoid[..., 0, :].reshape((-1, self.nb_landmarks, *z.shape))

            return np.stack([z_ellipsoid, y_ellipsoid, x_ellipsoid], axis=-1) + pred.reshape(
                (-1, self.nb_landmarks, 1, 1, self.spatial_dims)
            )
        else:
            raise ValueError("Unsupported spatial dimension")

    def in_region(self, pred, pred_cov, target, confidence=0.95):
        return mahalanobis_distance(pred, self.bound(pred_cov, confidence), target) <= 1

    def evaluate(self, pred, pred_cov, target, spacing=None, confidence=0.95, return_summary=True):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(pred, pred_cov)
        assert pred.shape == target.shape
        error = euclidean_distance(
            pixel_to_unit_numpy(pred, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = self.area_prediction_region(pred_cov, spacing=spacing, confidence=confidence)
        in_region = self.in_region(pred, pred_cov, target, confidence=confidence)
        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


class ConformalRegressorBonferroni:
    def __init__(
        self,
        spatial_dims=2,
        nb_landmarks=1,
    ):
        self.alphas = None
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks

    def check_input(self, pred):
        if pred.ndim != 3:
            raise ValueError("pred must have shape (N, C, D)")
        if pred.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        if pred.shape[2] != self.spatial_dims:
            raise ValueError("pred must have the same third dimension as spatial_dims")

    def fit(self, pred, target, sigmas=None):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(pred)
        if sigmas is None:
            sigmas = np.ones_like(target)
        elif isinstance(sigmas, torch.Tensor):
            sigmas = sigmas.detach().numpy()
        if sigmas.shape != target.shape:
            raise ValueError("sigmas must have the same shape as target")
        self.alphas = []
        for d in range(self.spatial_dims):
            alphas = np.abs(target[:, :, d] - pred[:, :, d]) / sigmas[:, :, d]
            alphas = np.sort(alphas, axis=0)[::-1]
            self.alphas.append(alphas)

    def bound(self, sigmas=None, confidence=0.95):
        """Get the bound for the given confidence level."""
        if self.alphas is None:
            raise ValueError("You must call fit before calling predict")
        if sigmas is None:
            sigmas = np.ones((self.spatial_dims,))
        bonferroni_confidence = 1 - (1 - confidence) / self.spatial_dims
        return (
            np.stack(
                [
                    alphas[int((1 - bonferroni_confidence) * (len(alphas) + 1)) - 1]
                    for alphas in self.alphas
                ],
                axis=-1,
            ).reshape((1, -1, self.spatial_dims))
            * sigmas
        )

    def area_prediction_region(self, sigmas=None, spacing=None, confidence=0.95):
        if spacing is None:
            return np.prod(self.bound(sigmas, confidence) * 2, axis=-1)
        return np.prod(self.bound(sigmas, confidence) * 2, axis=-1) * np.prod(
            spacing, axis=-1
        ).reshape(-1, 1)

    def predict(self, pred, sigma=None, confidence=0.95):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        return pred + self.bound(sigma, confidence) * np.array([-1, 1])

    def predict_contour(self, pred, sigmas, confidence=0.95, sample_points=200):
        # Convert torch tensor to numpy if needed
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()

        # Calculate error bounds - shape: (batch_size, nb_landmarks, spatial_dims)
        bounds = self.bound(sigmas, confidence)

        batch_size = pred.shape[0]

        if self.spatial_dims == 2:
            line_points = sample_points // 4

            # Initialize output arrays with the new shape
            x_coords = np.zeros((batch_size, line_points * 4, self.nb_landmarks))
            y_coords = np.zeros((batch_size, line_points * 4, self.nb_landmarks))

            # Generate contours for each batch and landmark
            for b in range(batch_size):
                for i in range(self.nb_landmarks):
                    # Get center point and bounds for current landmark
                    center_y = pred[b, i, 0]  # y coordinate
                    center_x = pred[b, i, 1]  # x coordinate
                    bound_y = bounds[b, i, 0]  # y bound
                    bound_x = bounds[b, i, 1]  # x bound

                    # Generate points for each edge
                    x_points = np.linspace(-bound_x, bound_x, line_points)
                    y_points = np.linspace(-bound_y, bound_y, line_points)

                    # Create rectangular contour
                    x_contour = np.concatenate(
                        [
                            center_x + x_points,  # Top edge
                            center_x + bound_x * np.ones(line_points),  # Right edge
                            center_x + x_points[::-1],  # Bottom edge
                            center_x - bound_x * np.ones(line_points),  # Left edge
                        ]
                    )

                    y_contour = np.concatenate(
                        [
                            center_y + bound_y * np.ones(line_points),  # Top edge
                            center_y + y_points,  # Right edge
                            center_y - bound_y * np.ones(line_points),  # Bottom edge
                            center_y + y_points[::-1],  # Left edge
                        ]
                    )

                    # Store the contour points in the new shape format
                    x_coords[b, :, i] = x_contour
                    y_coords[b, :, i] = y_contour

            return np.stack([y_coords, x_coords], axis=-1)

        elif self.spatial_dims == 3:
            bounds = self.bound(sigmas=sigmas, confidence=confidence)

            offsets = np.array(
                [
                    [-1, -1, -1],  # Vertex 0
                    [1, -1, -1],  # Vertex 1
                    [1, 1, -1],  # Vertex 2
                    [-1, 1, -1],  # Vertex 3
                    [-1, -1, 1],  # Vertex 4
                    [1, -1, 1],  # Vertex 5
                    [1, 1, 1],  # Vertex 6
                    [-1, 1, 1],  # Vertex 7
                ]
            )

            # Expand dimensions for broadcasting
            offsets = offsets[np.newaxis, np.newaxis, :, :]

            # Generate all vertices by scaling offsets with bounds and adding center
            vertices = pred[:, :, np.newaxis, :] + bounds[:, :, np.newaxis, :] * offsets

            # Define the face connectivity in terms of vertex indices
            face_indices = np.array(
                [
                    [0, 1, 5, 4],  # Bottom face
                    [2, 3, 7, 6],  # Top face
                    [0, 4, 7, 3],  # Left face
                    [1, 5, 6, 2],  # Right face
                    [0, 1, 2, 3],  # Front face
                    [4, 5, 6, 7],  # Back face
                ]
            )
            return vertices, face_indices
        else:
            raise ValueError("Unsupported spatial dimension")

    def in_region(self, pred, sigmas, target, confidence=0.95):
        return np.all(np.abs(target - pred) <= self.bound(sigmas, confidence), axis=-1)

    def evaluate(self, pred, sigmas, target, spacing=None, confidence=0.95, return_summary=True):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(sigmas, torch.Tensor):
            sigmas = sigmas.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(pred)
        assert pred.shape == target.shape
        error = euclidean_distance(
            pixel_to_unit_numpy(pred, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = self.area_prediction_region(sigmas, spacing=spacing, confidence=confidence)
        in_region = self.in_region(pred, sigmas, target, confidence=confidence)
        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


class ConformalRegressorMaxNonconformity:
    def __init__(
        self,
        spatial_dims=2,
        nb_landmarks=1,
    ):
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks
        self.alphas = None

    def check_input(self, pred):
        if pred.ndim != 3:
            raise ValueError("pred must have shape (N, C, D)")
        if pred.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        if pred.shape[2] != self.spatial_dims:
            raise ValueError("pred must have the same third dimension as spatial_dims")

    def fit(self, pred, target, sigmas=None):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(pred)
        if sigmas is None:
            sigmas = np.ones_like(target)
        elif isinstance(sigmas, torch.Tensor):
            sigmas = sigmas.detach().numpy()
        if sigmas.shape != target.shape:
            raise ValueError("sigmas must have the same shape as target")
        self.alphas = absolute_error_max(pred, target, sigmas)
        self.alphas = np.sort(self.alphas, axis=0)[::-1]

    def bound(self, sigmas=None, confidence=0.95):
        """Get the bound for the given confidence level."""
        if self.alphas is None:
            raise ValueError("You must call fit before calling predict")
        if sigmas is None:
            sigmas = np.ones((self.spatial_dims,))
        return (
            self.alphas[int((1 - confidence) * (len(self.alphas) + 1)) - 1].reshape((1, -1, 1))
            * sigmas
        )

    def area_prediction_region(self, sigmas=None, spacing=None, confidence=0.95):
        if spacing is None:
            return np.prod(self.bound(sigmas, confidence) * 2, axis=-1)
        return np.prod(self.bound(sigmas, confidence) * 2, axis=-1) * np.prod(
            spacing, axis=-1
        ).reshape(-1, 1)

    def predict(self, pred, sigma=None, confidence=0.95):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        return pred + self.bound(sigma, confidence) * np.array([-1, 1])

    def predict_contour(self, pred, sigmas, confidence=0.95, sample_points=200):
        # Convert torch tensor to numpy if needed
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()

        # Calculate error bounds - shape: (batch_size, nb_landmarks, spatial_dims)
        bounds = self.bound(sigmas, confidence)

        batch_size = pred.shape[0]

        if self.spatial_dims == 2:
            line_points = sample_points // 4

            # Initialize output arrays with the new shape
            x_coords = np.zeros((batch_size, line_points * 4, self.nb_landmarks))
            y_coords = np.zeros((batch_size, line_points * 4, self.nb_landmarks))

            # Generate contours for each batch and landmark
            for b in range(batch_size):
                for i in range(self.nb_landmarks):
                    # Get center point and bounds for current landmark
                    center_y = pred[b, i, 0]  # y coordinate
                    center_x = pred[b, i, 1]  # x coordinate
                    bound_y = bounds[b, i, 0]  # y bound
                    bound_x = bounds[b, i, 1]  # x bound

                    # Generate points for each edge
                    x_points = np.linspace(-bound_x, bound_x, line_points)
                    y_points = np.linspace(-bound_y, bound_y, line_points)

                    # Create rectangular contour
                    x_contour = np.concatenate(
                        [
                            center_x + x_points,  # Top edge
                            center_x + bound_x * np.ones(line_points),  # Right edge
                            center_x + x_points[::-1],  # Bottom edge
                            center_x - bound_x * np.ones(line_points),  # Left edge
                        ]
                    )

                    y_contour = np.concatenate(
                        [
                            center_y + bound_y * np.ones(line_points),  # Top edge
                            center_y + y_points,  # Right edge
                            center_y - bound_y * np.ones(line_points),  # Bottom edge
                            center_y + y_points[::-1],  # Left edge
                        ]
                    )

                    # Store the contour points in the new shape format
                    x_coords[b, :, i] = x_contour
                    y_coords[b, :, i] = y_contour

            return np.stack([y_coords, x_coords], axis=-1)

        elif self.spatial_dims == 3:
            bounds = self.bound(sigmas=sigmas, confidence=confidence)

            offsets = np.array(
                [
                    [-1, -1, -1],  # Vertex 0
                    [1, -1, -1],  # Vertex 1
                    [1, 1, -1],  # Vertex 2
                    [-1, 1, -1],  # Vertex 3
                    [-1, -1, 1],  # Vertex 4
                    [1, -1, 1],  # Vertex 5
                    [1, 1, 1],  # Vertex 6
                    [-1, 1, 1],  # Vertex 7
                ]
            )

            # Expand dimensions for broadcasting
            offsets = offsets[np.newaxis, np.newaxis, :, :]

            # Generate all vertices by scaling offsets with bounds and adding center
            vertices = pred[:, :, np.newaxis, :] + bounds[:, :, np.newaxis, :] * offsets

            # Define the face connectivity in terms of vertex indices
            face_indices = np.array(
                [
                    [0, 1, 5, 4],  # Bottom face
                    [2, 3, 7, 6],  # Top face
                    [0, 4, 7, 3],  # Left face
                    [1, 5, 6, 2],  # Right face
                    [0, 1, 2, 3],  # Front face
                    [4, 5, 6, 7],  # Back face
                ]
            )
            return vertices, face_indices
        else:
            raise ValueError("Unsupported spatial dimension")

    def in_region(self, pred, sigmas, target, confidence=0.95):
        return np.all(np.abs(target - pred) <= self.bound(sigmas, confidence), axis=-1)

    def evaluate(self, pred, sigmas, target, spacing=None, confidence=0.95, return_summary=True):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(sigmas, torch.Tensor):
            sigmas = sigmas.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(pred)
        assert pred.shape == target.shape
        error = euclidean_distance(
            pixel_to_unit_numpy(pred, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = self.area_prediction_region(sigmas, spacing=spacing, confidence=confidence)
        in_region = self.in_region(pred, sigmas, target, confidence=confidence)
        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


class MR2CCP:

    def __init__(
        self,
        spatial_dims=2,
        nb_landmarks=1,
        heatmap_size=(512, 512),
    ):
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks
        self.heatmap_size = heatmap_size
        self.alphas = None

    def check_input(self, heatmaps):
        if heatmaps.ndim != 4 and self.spatial_dims == 2:
            raise ValueError("pred must have shape (N, C, H, W)")
        elif heatmaps.ndim != 5 and self.spatial_dims == 3:
            raise ValueError("pred must have shape (N, C, D, H, W)")
        elif heatmaps.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        elif heatmaps.shape[2:] != self.heatmap_size:
            raise ValueError(
                f"pred {heatmaps.shape[2:]} must have the same dimensions as heatmap_size {self.heatmap_size}"
            )

    def fit(self, heatmaps, target, original_dims=None, padding=None):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(heatmaps)
        alphas = -self.conditional_probability(
            heatmaps=heatmaps, target=target, original_dims=original_dims, padding=padding
        )
        self.alphas = np.sort(alphas, axis=0)[::-1]

    def conditional_probability(self, heatmaps, target, original_dims=None, padding=None):
        resized_target = resize_landmarks(
            target, dim=np.array(self.heatmap_size), dim_orig=original_dims, padding=padding
        )
        # clip the target to the heatmap size
        resized_target = np.clip(resized_target, 0, np.array(self.heatmap_size) - 1)
        if self.spatial_dims == 2:
            y_min = np.floor(resized_target[..., 0]).astype(int)
            x_min = np.floor(resized_target[..., 1]).astype(int)
            y_max = np.ceil(resized_target[..., 0]).astype(int)
            x_max = np.ceil(resized_target[..., 1]).astype(int)

            prob_y_min_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                y_min.astype(int),
                x_min.astype(int),
            ]
            prob_y_min_x_max = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                y_min.astype(int),
                x_max.astype(int),
            ]
            prob_y_max_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                y_max.astype(int),
                x_min.astype(int),
            ]
            prob_y_max_x_max = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                y_max.astype(int),
                x_max.astype(int),
            ]

            diff_y_max = y_max - resized_target[..., 0]
            diff_x_max = x_max - resized_target[..., 1]
            diff_y_min = resized_target[..., 0] - y_min
            diff_x_min = resized_target[..., 1] - x_min

            # assign at indicecs y_max and y_min are the same diff_y_max and diff_y_min 0.5 to avoid division by zero
            idx_y_max_y_min = np.where(y_max == y_min)
            diff_y_max[idx_y_max_y_min] = 0.5
            diff_y_min[idx_y_max_y_min] = 0.5

            idx_x_max_x_min = np.where(x_max == x_min)
            diff_x_max[idx_x_max_x_min] = 0.5
            diff_x_min[idx_x_max_x_min] = 0.5

            weights_y_min_x_min = diff_y_max * diff_x_max
            weights_y_min_x_max = diff_y_max * diff_x_min
            weights_y_max_x_min = diff_y_min * diff_x_max
            weights_y_max_x_max = diff_y_min * diff_x_min

            prob = (
                prob_y_min_x_min * weights_y_min_x_min
                + prob_y_min_x_max * weights_y_min_x_max
                + prob_y_max_x_min * weights_y_max_x_min
                + prob_y_max_x_max * weights_y_max_x_max
            )
        elif self.spatial_dims == 3:
            z_min = np.floor(resized_target[..., 0])
            y_min = np.floor(resized_target[..., 1])
            x_min = np.floor(resized_target[..., 2])
            z_max = np.ceil(resized_target[..., 0])
            y_max = np.ceil(resized_target[..., 1])
            x_max = np.ceil(resized_target[..., 2])

            prob_z_min_y_min_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_min.astype(int),
                y_min.astype(int),
                x_min.astype(int),
            ]
            prob_z_min_y_min_x_max = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_min.astype(int),
                y_min.astype(int),
                x_max.astype(int),
            ]
            prob_z_min_y_max_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_min.astype(int),
                y_max.astype(int),
                x_min.astype(int),
            ]
            prob_z_min_y_max_x_max = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_min.astype(int),
                y_max.astype(int),
                x_max.astype(int),
            ]
            prob_z_max_y_min_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_max.astype(int),
                y_min.astype(int),
                x_min.astype(int),
            ]
            prob_z_max_y_min_x_max = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_max.astype(int),
                y_min.astype(int),
                x_max.astype(int),
            ]
            prob_z_max_y_max_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_max.astype(int),
                y_max.astype(int),
                x_min.astype(int),
            ]
            prob_z_max_y_max_x_max = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_max.astype(int),
                y_max.astype(int),
                x_max.astype(int),
            ]

            diff_z_max = z_max - resized_target[..., 0]
            diff_y_max = y_max - resized_target[..., 1]
            diff_x_max = x_max - resized_target[..., 2]
            diff_z_min = resized_target[..., 0] - z_min
            diff_y_min = resized_target[..., 1] - y_min
            diff_x_min = resized_target[..., 2] - x_min

            # assign at indicecs z_max and z_min are the same diff_z_max and diff_z_min 0.5 to avoid division by zero
            idx_z_max_z_min = np.where(z_max == z_min)
            diff_z_max[idx_z_max_z_min] = 0.5
            diff_z_min[idx_z_max_z_min] = 0.5

            idx_y_max_y_min = np.where(y_max == y_min)
            diff_y_max[idx_y_max_y_min] = 0.5
            diff_y_min[idx_y_max_y_min] = 0.5

            idx_x_max_x_min = np.where(x_max == x_min)
            diff_x_max[idx_x_max_x_min] = 0.5
            diff_x_min[idx_x_max_x_min] = 0.5

            weights_z_min_y_min_x_min = diff_z_max * diff_y_max * diff_x_max
            weights_z_min_y_min_x_max = diff_z_max * diff_y_max * diff_x_min
            weights_z_min_y_max_x_min = diff_z_max * diff_y_min * diff_x_max
            weights_z_min_y_max_x_max = diff_z_max * diff_y_min * diff_x_min
            weights_z_max_y_min_x_min = diff_z_min * diff_y_max * diff_x_max
            weights_z_max_y_min_x_max = diff_z_min * diff_y_max * diff_x_min
            weights_z_max_y_max_x_min = diff_z_min * diff_y_min * diff_x_max
            weights_z_max_y_max_x_max = diff_z_min * diff_y_min * diff_x_min
            prob = (
                prob_z_min_y_min_x_min * weights_z_min_y_min_x_min
                + prob_z_min_y_min_x_max * weights_z_min_y_min_x_max
                + prob_z_min_y_max_x_min * weights_z_min_y_max_x_min
                + prob_z_min_y_max_x_max * weights_z_min_y_max_x_max
                + prob_z_max_y_min_x_min * weights_z_max_y_min_x_min
                + prob_z_max_y_min_x_max * weights_z_max_y_min_x_max
                + prob_z_max_y_max_x_min * weights_z_max_y_max_x_min
                + prob_z_max_y_max_x_max * weights_z_max_y_max_x_max
            )
        return prob

    def bound(self, confidence=0.95):
        """Get the bound for the given confidence level."""
        if self.alphas is None:
            raise ValueError("You must call fit before calling predict")
        return self.alphas[int((1 - confidence) * (len(self.alphas) + 1)) - 1]

    def area_prediction_region(
        self, heatmaps, original_dims=None, padding=None, spacing=None, confidence=0.95
    ):
        area = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))
        print("Alert: area caluculation is a lower bound")
        for i in tqdm(range(heatmaps.shape[0])):
            pred_region = self.predict(
                heatmaps[i : i + 1],
                original_dims=original_dims[i : i + 1],
                padding=padding[i : i + 1],
                confidence=confidence,
            )
            if spacing is None:
                if self.spatial_dims == 2:
                    area[i] = np.sum(pred_region, axis=(-1, -2))
                elif self.spatial_dims == 3:
                    area[i] = np.sum(pred_region, axis=(-1, -2, -3))
            if self.spatial_dims == 2:
                area[i] = np.sum(pred_region, axis=(-1, -2)) * np.prod(spacing[i], axis=-1)
            elif self.spatial_dims == 3:
                area[i] = np.sum(pred_region, axis=(-1, -2, -3)) * np.prod(spacing[i], axis=-1)
        return area

    def predict(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        # Rescale the prediction region to the original dimensions
        mode = "bilinear" if self.spatial_dims == 2 else "trilinear"
        heatmaps = transform_heatmap_to_original_size_numpy(
            heatmaps, padding=padding, original_dim=original_dims, mode=mode
        )
        if self.spatial_dims == 2:
            return (
                -heatmaps <= self.bound(confidence).reshape((1, heatmaps.shape[1], 1, 1))
            ).astype(int)
        return (
            -heatmaps <= self.bound(confidence).reshape((1, heatmaps.shape[1], 1, 1, 1))
        ).astype(int)

    def predict_contour(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        print("Alert: contour caluculation is a lower bound")
        output = []
        for i in tqdm(range(heatmaps.shape[0])):
            if original_dims is None or padding is None:
                pred_region = self.predict(heatmaps[i : i + 1], confidence=confidence)[0]
            else:
                pred_region = self.predict(
                    heatmaps[i : i + 1],
                    original_dims=original_dims[i : i + 1],
                    padding=padding[i : i + 1],
                    confidence=confidence,
                )[0]
            output_sample = []
            for j in range(pred_region.shape[0]):
                pred_region_sample = pred_region[j]
                if self.spatial_dims == 2:
                    contour, _ = cv2.findContours(
                        pred_region_sample.astype(np.uint8),
                        cv2.RETR_EXTERNAL,
                        cv2.CHAIN_APPROX_SIMPLE,
                    )
                    output_sample.append(contour)
                elif self.spatial_dims == 3:
                    verts, faces, _, _ = measure.marching_cubes(
                        pred_region_sample, level=0.5
                    )  # returns verts, faces, normals, values
                    output_sample.append((verts, faces))
            output.append(output_sample)
        return output

    def in_region(
        self, heatmaps, target, original_dims=None, padding=None, confidence=0.95, batch_size=1
    ):
        return -self.conditional_probability(
            heatmaps=heatmaps, target=target, original_dims=original_dims, padding=padding
        ) <= self.bound(confidence)

    def evaluate(
        self,
        heatmaps,
        pred_landmarks,
        target,
        original_dims=None,
        padding=None,
        spacing=None,
        confidence=0.95,
        batch_size=1,
        return_summary=True,
    ):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(heatmaps)
        nb_batch = heatmaps.shape[0] // batch_size
        if heatmaps.shape[0] % batch_size != 0:
            nb_batch += 1
        error = euclidean_distance(
            pixel_to_unit_numpy(pred_landmarks, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))

        for i in tqdm(range(0, heatmaps.shape[0], batch_size)):
            batch_pred = heatmaps[i : i + batch_size]
            batch_original_dims = (
                original_dims[i : i + batch_size] if original_dims is not None else None
            )
            batch_padding = padding[i : i + batch_size] if padding is not None else None
            batch_spacing = spacing[i : i + batch_size] if spacing is not None else None

            batch_pred_region = self.predict(
                batch_pred,
                original_dims=batch_original_dims,
                padding=batch_padding,
                confidence=confidence,
            )
            if self.spatial_dims == 2:
                batch_area = np.sum(batch_pred_region, axis=(-1, -2)) * np.prod(
                    batch_spacing, axis=-1
                )
            elif self.spatial_dims == 3:
                batch_area = np.sum(batch_pred_region, axis=(-1, -2, -3)) * np.prod(
                    batch_spacing, axis=-1
                )

            area[i : i + batch_size] = batch_area
        in_region = self.in_region(heatmaps, target, original_dims, padding, confidence)
        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


class MR2C2R:

    def __init__(
        self,
        spatial_dims=2,
        nb_landmarks=1,
        heatmap_size=(512, 512),
        aps=False,
    ):
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks
        self.heatmap_size = heatmap_size
        self.aps = aps
        self.alphas = None

    def check_input(self, heatmaps):
        if heatmaps.ndim != 4 and self.spatial_dims == 2:
            raise ValueError("pred must have shape (N, C, H, W)")
        elif heatmaps.ndim != 5 and self.spatial_dims == 3:
            raise ValueError("pred must have shape (N, C, D, H, W)")
        elif heatmaps.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        elif heatmaps.shape[2:] != self.heatmap_size:
            raise ValueError(
                f"pred {heatmaps.shape[2:]} must have the same dimensions as heatmap_size {self.heatmap_size}"
            )

    def fit(self, heatmaps, target, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(heatmaps)
        if not self.aps:
            self.alphas = -self.conditional_probability(heatmaps, target, original_dims, padding)
            self.alphas = np.sort(self.alphas, axis=0)[::-1]
        else:
            self.fit_aps(heatmaps, target, original_dims, padding, confidence=confidence)

    def fit_aps(self, heatmaps, target, original_dims=None, padding=None, confidence=0.95):
        resized_target = resize_landmarks(
            target, dim=np.array(self.heatmap_size), dim_orig=original_dims, padding=padding
        )
        resized_target = np.floor(resized_target).astype(int)
        heatmaps_flatten = heatmaps.reshape((heatmaps.shape[0], heatmaps.shape[1], -1))
        if self.spatial_dims == 2:
            resized_target_flatten = (
                resized_target[..., 0] * heatmaps.shape[-1] + resized_target[..., 1]
            ).astype(int)
        else:
            resized_target_flatten = (
                resized_target[..., 0] * heatmaps.shape[-2] * heatmaps.shape[-1]
                + resized_target[..., 1] * heatmaps.shape[-1]
                + resized_target[..., 2]
            ).astype(int)
        idxs_sorted_heatmaps = np.argsort(heatmaps_flatten, axis=-1)[:, :, ::-1]
        sorted_heatmaps = np.take_along_axis(heatmaps_flatten, idxs_sorted_heatmaps, axis=-1)
        cum_sum_heatmaps = np.cumsum(sorted_heatmaps, axis=-1)

        if self.alphas is None:
            self.alphas = {}
        elif not isinstance(self.alphas, dict):
            raise ValueError("Method already fitted with non-aps method")
        self.alphas[confidence] = []
        for j in range(heatmaps_flatten.shape[1]):
            # map the target to the index of the sorted heatmaps
            idx_target = np.where(resized_target_flatten[:, j, None] == idxs_sorted_heatmaps[:, j])[
                1
            ]
            cum_sum_target = cum_sum_heatmaps[np.arange(cum_sum_heatmaps.shape[0]), j, idx_target]
            scores = cum_sum_target - confidence
            level_adjusted = confidence * (1 + 1 / len(scores))
            correction = mquantiles(scores, prob=level_adjusted)
            corrected_confidence = np.minimum(confidence + correction, 1.0)
            self.alphas[confidence].append(corrected_confidence)

    def conditional_probability(self, heatmaps, target, original_dims=None, padding=None):
        resized_target = resize_landmarks(
            target, dim=np.array(self.heatmap_size), dim_orig=original_dims, padding=padding
        )
        min_idxs = np.floor(resized_target).astype(int)

        if self.spatial_dims == 2:
            y_min = min_idxs[:, :, 0]
            x_min = min_idxs[:, :, 1]
            prob_y_min_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None], np.arange(heatmaps.shape[1]), y_min, x_min
            ]
            prob = prob_y_min_x_min

        elif self.spatial_dims == 3:
            z_min = min_idxs[:, :, 0]
            y_min = min_idxs[:, :, 1]
            x_min = min_idxs[:, :, 2]
            prob_z_min_y_min_x_min = heatmaps[
                np.arange(heatmaps.shape[0])[:, None],
                np.arange(heatmaps.shape[1]),
                z_min,
                y_min,
                x_min,
            ]
            prob = prob_z_min_y_min_x_min

        else:
            raise ValueError("Unsupported spatial dimension")

        return prob

    def bound(self, confidence=0.95):
        """Get the bound for the given confidence level."""
        if self.alphas is None:
            raise ValueError("You must call fit before calling predict")
        elif not self.aps:
            return self.alphas[int((1 - confidence) * (len(self.alphas) + 1)) - 1]
        elif confidence not in self.alphas.keys():
            raise ValueError("Confidence level not fitted")
        return self.alphas[confidence]

    def area_prediction_region(
        self, heatmaps, original_dims=None, padding=None, spacing=None, confidence=0.95
    ):
        area = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))
        for i in tqdm(range(heatmaps.shape[0])):
            pred_region = self.predict(
                heatmaps[i : i + 1],
                original_dims=original_dims[i : i + 1],
                padding=padding[i : i + 1],
                confidence=confidence,
            )
            if spacing is None:
                if self.spatial_dims == 2:
                    area[i] = np.sum(pred_region, axis=(-1, -2))
                elif self.spatial_dims == 3:
                    area[i] = np.sum(pred_region, axis=(-1, -2, -3))
            if self.spatial_dims == 2:
                area[i] = np.sum(pred_region, axis=(-1, -2)) * np.prod(spacing[i], axis=-1)
            elif self.spatial_dims == 3:
                area[i] = np.sum(pred_region, axis=(-1, -2, -3)) * np.prod(spacing[i], axis=-1)
        return area

    def predict(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if self.aps:
            return self.predict_aps(heatmaps, original_dims, padding, confidence)
        return self.predict_non_aps(heatmaps, original_dims, padding, confidence)

    def predict_non_aps(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        if original_dims is None:
            if self.spatial_dims == 2:
                return (
                    -heatmaps <= self.bound(confidence).reshape((1, heatmaps.shape[1], 1, 1))
                ).astype(int)
            elif self.spatial_dims == 3:
                return (
                    -heatmaps <= self.bound(confidence).reshape((1, heatmaps.shape[1], 1, 1, 1))
                ).astype(int)
        if padding is None:
            padding = np.zeros((heatmaps.shape[0], self.spatial_dims))
        if self.spatial_dims == 2:
            pred_region = -heatmaps <= self.bound(confidence).reshape((1, heatmaps.shape[1], 1, 1))
        elif self.spatial_dims == 3:
            pred_region = -heatmaps <= self.bound(confidence).reshape(
                (1, heatmaps.shape[1], 1, 1, 1)
            )
        # Rescale the prediction region to the original dimensions
        mode = "nearest-exact"
        pred_region = transform_heatmap_to_original_size_numpy(
            pred_region.astype(float), padding=padding, original_dim=original_dims, mode=mode
        )
        return (pred_region > 0).astype(int)

    def predict_aps(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        corrected_confidences = self.bound(confidence)
        # clone heatmaps
        heatmaps = np.copy(heatmaps)
        for i in range(heatmaps.shape[0]):
            for j in range(heatmaps.shape[1]):
                idxs_sorted_heatmaps = np.argsort(heatmaps[i, j].flatten())[::-1]
                cum_sum_sorted_heatmaps = np.cumsum(heatmaps[i, j].flatten()[idxs_sorted_heatmaps])
                if corrected_confidences[j] >= len(cum_sum_sorted_heatmaps) / (
                    len(cum_sum_sorted_heatmaps) + 1
                ):
                    threshold = len(cum_sum_sorted_heatmaps)
                else:
                    threshold = np.argmax(cum_sum_sorted_heatmaps >= corrected_confidences[j])
                heatmaps[i, j] = np.zeros_like(heatmaps[i, j])
                flatten_heatmaps = heatmaps[i, j].flatten()
                flatten_heatmaps[idxs_sorted_heatmaps[:threshold]] = 1
                heatmaps[i, j] = flatten_heatmaps.reshape(heatmaps[i, j].shape)

        if original_dims is None:
            return heatmaps
        # Rescale the prediction region to the original dimensions
        mode = "nearest-exact"
        heatmap = transform_heatmap_to_original_size_numpy(
            heatmaps.astype(float), padding=padding, original_dim=original_dims, mode=mode
        )
        return (heatmap > 0).astype(int)

    def predict_contour(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        output = []
        for i in tqdm(range(heatmaps.shape[0])):
            if original_dims is None or padding is None:
                pred_region = self.predict(heatmaps[i : i + 1], confidence=confidence)[0]
            else:
                pred_region = self.predict(
                    heatmaps[i : i + 1],
                    original_dims=original_dims[i : i + 1],
                    padding=padding[i : i + 1],
                    confidence=confidence,
                )[0]
            output_sample = []
            for j in range(pred_region.shape[0]):
                pred_region_sample = pred_region[j]
                if self.spatial_dims == 2:
                    contour, _ = cv2.findContours(
                        pred_region_sample.astype(np.uint8),
                        cv2.RETR_EXTERNAL,
                        cv2.CHAIN_APPROX_SIMPLE,
                    )
                    output_sample.append(contour)
                elif self.spatial_dims == 3:
                    verts, faces, _, _ = measure.marching_cubes(
                        pred_region_sample, level=0.5
                    )  # returns verts, faces, normals, values
                    output_sample.append((verts, faces))
            output.append(output_sample)
        return output

    def in_region(
        self, heatmaps, target, original_dims=None, padding=None, confidence=0.95, batch_size=1
    ):
        output = []
        for i in tqdm(range(0, heatmaps.shape[0], batch_size)):
            pred_region = self.predict(
                heatmaps[i : i + 1],
                original_dims=original_dims[i : i + 1],
                padding=padding[i : i + 1],
                confidence=confidence,
            )
            if self.spatial_dims == 2:
                target_y_min = np.floor(target[i : i + 1, :, 0]).astype(int)
                target_x_min = np.floor(target[i : i + 1, :, 1]).astype(int)
                in_y_min_x_min = pred_region[
                    np.arange(pred_region.shape[0])[:, None],
                    np.arange(pred_region.shape[1]),
                    target_y_min,
                    target_x_min,
                ]
            elif self.spatial_dims == 3:
                target_z_min = np.floor(target[i : i + 1, :, 0]).astype(int)
                target_y_min = np.floor(target[i : i + 1, :, 1]).astype(int)
                target_x_min = np.floor(target[i : i + 1, :, 2]).astype(int)
                in_y_min_x_min = pred_region[
                    np.arange(pred_region.shape[0])[:, None],
                    np.arange(pred_region.shape[1]),
                    target_z_min,
                    target_y_min,
                    target_x_min,
                ]
            output.append(in_y_min_x_min)
        return np.concatenate(output, axis=0)

    def evaluate(
        self,
        heatmaps,
        pred_landmarks,
        target,
        original_dims=None,
        padding=None,
        spacing=None,
        confidence=0.95,
        batch_size=1,
        return_summary=True,
    ):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(heatmaps)
        nb_batch = heatmaps.shape[0] // batch_size
        if heatmaps.shape[0] % batch_size != 0:
            nb_batch += 1
        error = euclidean_distance(
            pixel_to_unit_numpy(pred_landmarks, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))
        in_region = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))

        for i in tqdm(range(0, heatmaps.shape[0], batch_size)):
            batch_pred = heatmaps[i : i + batch_size]
            batch_target = target[i : i + batch_size]
            batch_original_dims = (
                original_dims[i : i + batch_size] if original_dims is not None else None
            )
            batch_padding = padding[i : i + batch_size] if padding is not None else None
            batch_spacing = spacing[i : i + batch_size] if spacing is not None else None

            batch_pred_region = self.predict(
                batch_pred,
                original_dims=batch_original_dims,
                padding=batch_padding,
                confidence=confidence,
            )
            if self.spatial_dims == 2:
                batch_area = np.sum(batch_pred_region, axis=(-1, -2)) * np.prod(
                    batch_spacing, axis=-1
                )
                batch_target_y_min = np.floor(batch_target[:, :, 0]).astype(int)
                batch_target_x_min = np.floor(batch_target[:, :, 1]).astype(int)

                in_y_min_x_min = batch_pred_region[
                    np.arange(batch_pred_region.shape[0])[:, None],
                    np.arange(batch_pred_region.shape[1]),
                    batch_target_y_min,
                    batch_target_x_min,
                ]
            elif self.spatial_dims == 3:
                batch_area = np.sum(batch_pred_region, axis=(-1, -2, -3)) * np.prod(
                    batch_spacing, axis=-1
                )
                batch_target_z_min = np.floor(batch_target[:, :, 0]).astype(int)
                batch_target_y_min = np.floor(batch_target[:, :, 1]).astype(int)
                batch_target_x_min = np.floor(batch_target[:, :, 2]).astype(int)

                in_y_min_x_min = batch_pred_region[
                    np.arange(batch_pred_region.shape[0])[:, None],
                    np.arange(batch_pred_region.shape[1]),
                    batch_target_z_min,
                    batch_target_y_min,
                    batch_target_x_min,
                ]

            batch_coverage = in_y_min_x_min
            area[i : i + batch_size] = batch_area
            in_region[i : i + batch_size] = batch_coverage

        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


class MultivariateNormalRegressor:
    def __init__(self, spatial_dims=2, nb_landmarks=1):
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks

    def check_input(self, pred, pred_cov):
        if pred.ndim != 3:
            raise ValueError("pred must have shape (N, C, D)")
        if pred_cov.ndim != 4:
            raise ValueError("pred_cov must have shape (N, C, D, D)")
        if pred.shape[2] != pred_cov.shape[2] and self.spatial_dims != pred.shape[2]:
            raise ValueError("pred and pred_cov must have the same second dimension")
        if pred_cov.shape[2] != pred_cov.shape[3]:
            raise ValueError("pred_cov must be square")
        if pred.shape[0] != pred_cov.shape[0]:
            raise ValueError("pred and pred_cov must have the same first dimension")
        if pred.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        if pred_cov.shape[1] != self.nb_landmarks:
            raise ValueError("pred_cov must have the same second dimension as nb_landmarks")

    def fit(self, pred, pred_cov, target):
        return

    def bound(self, pred_cov, confidence=0.95):
        """Get the bound for the given confidence level."""
        return pred_cov * norm.ppf(confidence) ** 2

    def area_prediction_region(self, pred_cov, spacing=None, confidence=0.95):
        """
        Calculate the area of the prediction region, accounting for pixel spacing.

        Parameters:
        pred_cov (ndarray): The predicted covariance matrix.
        spacing (ndarray, optional): The spacing between pixels in millimeters,
                         with shape (Batch_size, spatial_size).
                         Defaults to None.
        confidence (float, optional): The confidence level for the prediction region.
                         Defaults to 0.95.

        Returns:
        float: The area of the prediction region.
        """
        pred_cov_scaled = self.bound(pred_cov, confidence)
        if spacing is not None:
            if isinstance(spacing, torch.Tensor):
                spacing = spacing.detach().numpy()
            assert spacing.shape == (pred_cov.shape[0], self.spatial_dims)
            # Make a diagonal matrix with the pixel spacing (Batch_size, spatial_size, spatial_size)
            # e.g. for 2D: [[s1, 0], [0, s2]], for 3D: [[s1, 0, 0], [0, s2, 0], [0, 0, s3]]
            spacing_matrix = np.stack(
                [np.diag(spacing[i]) for i in range(spacing.shape[0])], axis=0
            ).reshape((pred_cov.shape[0], 1, self.spatial_dims, self.spatial_dims))
            # Scale the covariance matrix by the pixel spacing
            pred_cov_scaled = (
                spacing_matrix @ pred_cov_scaled @ np.transpose(spacing_matrix, (0, 1, 3, 2))
            )
        eig = np.linalg.eigvalsh(pred_cov_scaled)
        return np.pi * np.prod(np.sqrt(eig), axis=-1)

    def predict(self, pred, pred_cov, confidence=0.95):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        self.check_input(pred, pred_cov)
        return pred, self.bound(pred_cov, confidence)

    def predict_contour(self, pred, pred_cov, confidence=0.95, sample_points=200):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        self.check_input(pred, pred_cov)

        if self.spatial_dims == 2:
            angles = np.linspace(0, 2 * np.pi, sample_points)
            directions = np.stack((np.cos(angles), np.sin(angles)), axis=-1)
            directions = directions.reshape((1, -1, 1, 1, self.spatial_dims))
            error_bound = directions @ np.linalg.cholesky(self.bound(pred_cov, confidence)).reshape(
                (-1, 1, self.nb_landmarks, self.spatial_dims, self.spatial_dims)
            )
            return pred.reshape(
                (-1, 1, self.nb_landmarks, self.spatial_dims)
            ) + error_bound.reshape((-1, directions.shape[1], self.nb_landmarks, self.spatial_dims))
        elif self.spatial_dims == 3:
            eigenvalues, eigenvectors = np.linalg.eigh(self.bound(pred_cov, confidence))

            # Create a grid of points for a unit sphere
            u = np.linspace(0, 2 * np.pi, sample_points)
            v = np.linspace(0, np.pi, sample_points // 2)
            x = np.outer(np.cos(u), np.sin(v))
            y = np.outer(np.sin(u), np.sin(v))
            z = np.outer(np.ones_like(u), np.cos(v))

            # Scale and rotate the unit sphere to match the ellipsoid
            radii = np.sqrt(eigenvalues)
            transform = eigenvectors @ np.stack(
                [
                    np.stack(
                        [radii[..., 0], np.zeros_like(radii[..., 0]), np.zeros_like(radii[..., 0])],
                        axis=-1,
                    ),
                    np.stack(
                        [np.zeros_like(radii[..., 0]), radii[..., 1], np.zeros_like(radii[..., 0])],
                        axis=-1,
                    ),
                    np.stack(
                        [np.zeros_like(radii[..., 0]), np.zeros_like(radii[..., 0]), radii[..., 2]],
                        axis=-1,
                    ),
                ],
                axis=-2,
            )
            ellipsoid = transform @ np.array([z.ravel(), y.ravel(), x.ravel()])
            # Reshape to 3D grid
            x_ellipsoid = ellipsoid[..., 2, :].reshape((-1, self.nb_landmarks, *x.shape))
            y_ellipsoid = ellipsoid[..., 1, :].reshape((-1, self.nb_landmarks, *y.shape))
            z_ellipsoid = ellipsoid[..., 0, :].reshape((-1, self.nb_landmarks, *z.shape))

            return np.stack([z_ellipsoid, y_ellipsoid, x_ellipsoid], axis=-1) + pred.reshape(
                (-1, self.nb_landmarks, 1, 1, self.spatial_dims)
            )
        else:
            raise ValueError("Unsupported spatial dimension")

    def in_region(self, pred, pred_cov, target, confidence=0.95):
        return mahalanobis_distance(pred, self.bound(pred_cov, confidence), target) <= 1

    def evaluate(self, pred, pred_cov, target, spacing=None, confidence=0.95, return_summary=True):
        if isinstance(pred, torch.Tensor):
            pred = pred.detach().numpy()
        if isinstance(pred_cov, torch.Tensor):
            pred_cov = pred_cov.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(pred, pred_cov)
        assert pred.shape == target.shape
        error = euclidean_distance(
            pixel_to_unit_numpy(pred, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = self.area_prediction_region(pred_cov, spacing=spacing, confidence=confidence)
        in_region = self.in_region(pred, pred_cov, target, confidence=confidence)
        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


class ContourHuggingRegressor:

    def __init__(self, spatial_dims=2, nb_landmarks=1, heatmap_size=(512, 512)):
        self.spatial_dims = spatial_dims
        self.nb_landmarks = nb_landmarks
        self.heatmap_size = heatmap_size

    def check_input(self, heatmaps):
        if heatmaps.ndim != 4 and self.spatial_dims == 2:
            raise ValueError("pred must have shape (N, C, H, W)")
        elif heatmaps.ndim != 5 and self.spatial_dims == 3:
            raise ValueError("pred must have shape (N, C, D, H, W)")
        elif heatmaps.shape[1] != self.nb_landmarks:
            raise ValueError("pred must have the same second dimension as nb_landmarks")
        elif heatmaps.shape[2:] != self.heatmap_size:
            raise ValueError(
                f"pred {heatmaps.shape[2:]} must have the same dimensions as heatmap_size {self.heatmap_size}"
            )

    def fit(self, heatmaps, target, original_dims=None, padding=None):
        return

    def area_prediction_region(
        self, heatmaps, original_dims=None, padding=None, spacing=None, confidence=0.95
    ):
        area = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))
        for i in tqdm(range(heatmaps.shape[0])):
            pred_region = self.predict(
                heatmaps[i : i + 1],
                original_dims=original_dims[i : i + 1],
                padding=padding[i : i + 1],
                confidence=confidence,
            )
            if spacing is None:
                if self.spatial_dims == 2:
                    area[i] = np.sum(pred_region, axis=(-1, -2))
                elif self.spatial_dims == 3:
                    area[i] = np.sum(pred_region, axis=(-1, -2, -3))
            if self.spatial_dims == 2:
                area[i] = np.sum(pred_region, axis=(-1, -2)) * np.prod(spacing[i], axis=-1)
            elif self.spatial_dims == 3:
                area[i] = np.sum(pred_region, axis=(-1, -2, -3)) * np.prod(spacing[i], axis=-1)
        return area

    def predict(self, heatmaps, original_dims=None, padding=None, spacing=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        # clone heatmaps
        heatmaps = np.copy(heatmaps)
        for i in range(heatmaps.shape[0]):
            for j in range(heatmaps.shape[1]):
                idxs_sorted_heatmaps = np.argsort(heatmaps[i, j].flatten())[::-1]
                cum_sum_sorted_heatmaps = np.cumsum(heatmaps[i, j].flatten()[idxs_sorted_heatmaps])
                threshold = np.argmax(cum_sum_sorted_heatmaps >= confidence)
                heatmaps[i, j] = np.zeros_like(heatmaps[i, j])
                flatten_heatmaps = heatmaps[i, j].flatten()
                flatten_heatmaps[idxs_sorted_heatmaps[:threshold]] = 1
                heatmaps[i, j] = flatten_heatmaps.reshape(heatmaps[i, j].shape)
        if original_dims is None:
            return heatmaps
        # Rescale the prediction region to the original dimensions
        heatmap = transform_heatmap_to_original_size_numpy(
            heatmaps.astype(float), padding=padding, original_dim=original_dims
        )
        return (heatmap > 0).astype(int)

    def predict_contour(self, heatmaps, original_dims=None, padding=None, confidence=0.95):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()

        output = []
        for i in tqdm(range(heatmaps.shape[0])):
            if original_dims is None or padding is None:
                pred_region = self.predict(heatmaps[i : i + 1], confidence=confidence)[0]
            else:
                pred_region = self.predict(
                    heatmaps[i : i + 1],
                    original_dims=original_dims[i : i + 1],
                    padding=padding[i : i + 1],
                    confidence=confidence,
                )[0]
            output_sample = []
            for j in range(pred_region.shape[0]):
                pred_region_sample = pred_region[j]
                if self.spatial_dims == 2:
                    contour, _ = cv2.findContours(
                        pred_region_sample.astype(np.uint8),
                        cv2.RETR_EXTERNAL,
                        cv2.CHAIN_APPROX_SIMPLE,
                    )
                    output_sample.append(contour)
                elif self.spatial_dims == 3:
                    verts, faces, _, _ = measure.marching_cubes(
                        pred_region_sample, level=0.5
                    )  # returns verts, faces, normals, values
                    output_sample.append((verts, faces))
            output.append(output_sample)
        return output

    def in_region(
        self, heatmaps, target, original_dims=None, padding=None, confidence=0.95, batch_size=1
    ):
        output = []
        for i in tqdm(range(0, heatmaps.shape[0], batch_size)):
            pred_region = self.predict(
                heatmaps[i : i + 1],
                original_dims=original_dims[i : i + 1],
                padding=padding[i : i + 1],
                confidence=confidence,
            )
            if self.spatial_dims == 2:
                target_y_min = np.floor(target[i : i + 1, :, 0]).astype(int)
                target_x_min = np.floor(target[i : i + 1, :, 1]).astype(int)
                in_y_min_x_min = pred_region[
                    np.arange(pred_region.shape[0])[:, None],
                    np.arange(pred_region.shape[1]),
                    target_y_min,
                    target_x_min,
                ]
            elif self.spatial_dims == 3:
                target_z_min = np.floor(target[i : i + 1, :, 0]).astype(int)
                target_y_min = np.floor(target[i : i + 1, :, 1]).astype(int)
                target_x_min = np.floor(target[i : i + 1, :, 2]).astype(int)
                in_y_min_x_min = pred_region[
                    np.arange(pred_region.shape[0])[:, None],
                    np.arange(pred_region.shape[1]),
                    target_z_min,
                    target_y_min,
                    target_x_min,
                ]
            output.append(in_y_min_x_min)
        return np.concatenate(output, axis=0)

    def evaluate(
        self,
        heatmaps,
        pred_landmarks,
        target,
        original_dims=None,
        padding=None,
        spacing=None,
        confidence=0.95,
        batch_size=1,
        return_summary=True,
    ):
        if isinstance(heatmaps, torch.Tensor):
            heatmaps = heatmaps.detach().numpy()
        if isinstance(target, torch.Tensor):
            target = target.detach().numpy()
        self.check_input(heatmaps)
        nb_batch = heatmaps.shape[0] // batch_size
        if heatmaps.shape[0] % batch_size != 0:
            nb_batch += 1
        error = euclidean_distance(
            pixel_to_unit_numpy(pred_landmarks, pixel_spacing=spacing),
            pixel_to_unit_numpy(target, pixel_spacing=spacing),
        )
        area = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))
        in_region = np.zeros((heatmaps.shape[0], heatmaps.shape[1]))

        for i in tqdm(range(0, heatmaps.shape[0], batch_size)):
            batch_pred = heatmaps[i : i + batch_size]
            batch_target = target[i : i + batch_size]
            batch_original_dims = (
                original_dims[i : i + batch_size] if original_dims is not None else None
            )
            batch_padding = padding[i : i + batch_size] if padding is not None else None
            batch_spacing = spacing[i : i + batch_size] if spacing is not None else None

            batch_pred_region = self.predict(
                batch_pred,
                original_dims=batch_original_dims,
                padding=batch_padding,
                confidence=confidence,
            )
            if self.spatial_dims == 2:
                batch_area = np.sum(batch_pred_region, axis=(-1, -2)) * np.prod(
                    batch_spacing, axis=-1
                )
                batch_target_y_min = np.floor(batch_target[:, :, 0]).astype(int)
                batch_target_x_min = np.floor(batch_target[:, :, 1]).astype(int)

                in_y_min_x_min = batch_pred_region[
                    np.arange(batch_pred_region.shape[0])[:, None],
                    np.arange(batch_pred_region.shape[1]),
                    batch_target_y_min,
                    batch_target_x_min,
                ]
            elif self.spatial_dims == 3:
                batch_area = np.sum(batch_pred_region, axis=(-1, -2, -3)) * np.prod(
                    batch_spacing, axis=-1
                )
                batch_target_z_min = np.floor(batch_target[:, :, 0]).astype(int)
                batch_target_y_min = np.floor(batch_target[:, :, 1]).astype(int)
                batch_target_x_min = np.floor(batch_target[:, :, 2]).astype(int)

                in_y_min_x_min = batch_pred_region[
                    np.arange(batch_pred_region.shape[0])[:, None],
                    np.arange(batch_pred_region.shape[1]),
                    batch_target_z_min,
                    batch_target_y_min,
                    batch_target_x_min,
                ]

            batch_coverage = in_y_min_x_min
            area[i : i + batch_size] = batch_area
            in_region[i : i + batch_size] = batch_coverage

        coverage = in_region.mean()
        efficiency_mean = area.mean()
        efficiency_median = np.median(area)
        efficiency_q1 = np.quantile(area, 0.25)
        efficiency_q3 = np.quantile(area, 0.75)
        adaptivity = spearmanr(error.flatten(), area.flatten()).statistic
        print(f"Coverage: {coverage}")
        print(f"Efficiency (mean): {efficiency_mean}")
        print(f"Efficiency (median): {efficiency_median}")
        print(f"Efficiency (Q1): {efficiency_q1}")
        print(f"Efficiency (Q3): {efficiency_q3}")
        print(f"Adeptivity: {adaptivity}")
        if not return_summary:
            return (in_region, area, error)
        return {
            "coverage": coverage,
            "efficiency_mean": efficiency_mean,
            "efficiency_median": efficiency_median,
            "efficiency_q1": efficiency_q1,
            "efficiency_q3": efficiency_q3,
            "adaptivity": adaptivity,
        }


def transform_heatmap_to_original_size_numpy(heatmaps, padding, original_dim, mode="nearest-exact"):
    heatmap = torch.tensor(heatmaps)
    padding = torch.tensor(padding)
    original_dim = torch.tensor(original_dim)
    return (
        transform_heatmap_to_original_size(heatmap, padding, original_dim, mode=mode)
        .detach()
        .numpy()
    )


def transform_heatmap_to_original_size(heatmaps, padding, original_dim, mode="nearest-exact"):
    heatmaps_orig_size = []
    for i in range(heatmaps.shape[0]):
        # resize image to original size
        resize_dim = original_dim[i] + 2 * padding[i]
        resize_dim = tuple(int(dim.item()) for dim in resize_dim)
        heatmap = torch.nn.functional.interpolate(
            heatmaps[i].unsqueeze(0), size=resize_dim, mode=mode
        )
        # crop image to original size
        if heatmaps.ndim == 4:  # 2D case
            heatmaps_orig_size.append(
                heatmap[
                    :,
                    :,
                    int(padding[i, 0]) : int(padding[i, 0] + original_dim[i, 0].item()),
                    int(padding[i, 1]) : int(padding[i, 1] + original_dim[i, 1].item()),
                ]
            )
        elif heatmaps.ndim == 5:  # 3D case
            heatmaps_orig_size.append(
                heatmap[
                    :,
                    :,
                    int(padding[i, 0]) : int(padding[i, 0] + original_dim[i, 0].item()),
                    int(padding[i, 1]) : int(padding[i, 1] + original_dim[i, 1].item()),
                    int(padding[i, 2]) : int(padding[i, 2] + original_dim[i, 2].item()),
                ]
            )
    return torch.cat(heatmaps_orig_size, dim=0)


def resize_landmarks(
    landmarks: np.ndarray,
    dim: Optional[np.ndarray] = None,
    dim_orig: Optional[np.ndarray] = None,
    padding: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Convert the landmarks from pixel to unit (Numpy version).

    Args:
        landmarks (np.ndarray): landmarks
        pixel_spacing (Optional[np.ndarray], optional): pixel spacing. Defaults to None.
        dim (Optional[tuple[int, ...] | np.ndarray], optional): image size. Defaults to None.
        dim_orig (Optional[np.ndarray], optional): original image size. Defaults to None.
        padding (Optional[np.ndarray], optional): padding. Defaults to None.

    Returns:
        np.ndarray: landmarks in units
    """
    spatial_dim = landmarks.shape[-1]

    if dim is not None:
        assert dim.shape[-1] == spatial_dim, f"dim must have {spatial_dim} elements."
    if dim_orig is not None:
        assert dim_orig.shape[-1] == spatial_dim, f"dim_orig must have {spatial_dim} elements."
    if padding is None:
        padding = np.zeros_like(landmarks)
    t_landmarks = landmarks + padding.reshape((landmarks.shape[0], 1, spatial_dim))
    t_landmarks = t_landmarks * (dim / (dim_orig + 2 * padding)).reshape((-1, 1, spatial_dim))  # type: ignore
    return t_landmarks



"""
Utils module.
"""

from .preprocessing import extract_roi, normalize, preprocess_all
from .utils import (
    all_annotations_to_landmarks,
    all_annotations_to_landmarks_numpy,
    annotation_to_landmark,
    annotation_to_landmark_numpy,
    covert_video_to_frames,
    get_angle,
    get_angle_numpy,
    get_paths,
    pixel_to_unit,
    pixel_to_unit_numpy,
)

__all__ = [
    "append_files_extension",
    "convert_all_dcm_png",
    "normalize",
    "preprocess_all",
    "get_paths",
    "annotation_to_landmark",
    "annotation_to_landmark_numpy",
    "all_annotations_to_landmarks",
    "all_annotations_to_landmarks_numpy",
    "get_angle",
    "get_angle_numpy",
    "pixel_to_unit",
    "pixel_to_unit_numpy",
    "covert_video_to_frames",
    "extract_roi",
]



"""
Multi-instance and multi-class landmark dataset using data from the EndoVis 2015 challenge. The
dataset contains 4 training and 6 testing videos of robotic surgery. The goal is to predict the
location of instruments in the video, more specifically the tip of the clasper. We only consider the
clasper points and ignore the other points, since they are way more ambiguous. The videos are
transformed into images and the annotations are given as 2D points. The dataset is split into a
training and testing set. The training set contains 4 videos and the testing set contains 6 videos,
such as specified in the challenge. However, it is possible to use the other points by specifying
the class_names argument.

The dataset is available at https://opencas.webarchiv.kit.edu/data/endovis15_ins and the annotations
are available at https://github.com/surgical-vision/EndoVisPoseAnnotation.
"""

import json
import os
import zipfile

import opendatasets as od  # type: ignore
import torch

from landmarker.data import HeatmapDataset, LandmarkDataset
from landmarker.utils import covert_video_to_frames

# class_names = ["LeftClasperPoint", "RightClasperPoint", "HeadPoint",
#                "ShaftPoint", "TrackedPoint", "EndPoint"]


def get_endovis2015_dataset(
    path_dir: str, class_names: list[str] = ["LeftClasperPoint", "RightClasperPoint"]
):
    if not os.path.exists(path_dir + "/EndoVis2015"):
        od.download_url(
            "https://opencas.webarchiv.kit.edu/data/endovis15_ins/Tracking_Robotic_Training.zip",
            path_dir,
        )
        od.download_url(
            "https://opencas.webarchiv.kit.edu/data/endovis15_ins/Tracking_Robotic_Testing.zip",
            path_dir,
        )
        with zipfile.ZipFile(path_dir + "/Tracking_Robotic_Training.zip", "r") as zip_ref:
            zip_ref.extractall(path=path_dir + "/EndoVis2015")
        with zipfile.ZipFile(path_dir + "/Tracking_Robotic_Testing.zip", "r") as zip_ref:
            zip_ref.extractall(path=path_dir + "/EndoVis2015")
        os.rename(
            path_dir + "/EndoVis2015/Tracking_Robotic_Training/Training",
            path_dir + "/EndoVis2015/Training",
        )
        os.rmdir(path_dir + "/EndoVis2015/Tracking_Robotic_Training")
        os.rename(
            path_dir + "/EndoVis2015/Training", path_dir + "/EndoVis2015/Tracking_Robotic_Training"
        )
        os.rename(
            path_dir + "/EndoVis2015/Tracking", path_dir + "/EndoVis2015/Tracking_Robotic_Testing"
        )
        os.remove(path_dir + "/Tracking_Robotic_Training.zip")
        os.remove(path_dir + "/Tracking_Robotic_Testing.zip")
        label_path = (
            "https://raw.githubusercontent.com/" + "surgical-vision/EndoVisPoseAnnotation/master"
        )
        for i in range(4):
            od.download_url(
                f"{label_path}/train_labels/train{i+1}_labels.json",
                path_dir + "/EndoVis2015/Tracking_Robotic_Training/labels",
            )
        for i in range(6):
            od.download_url(
                f"{label_path}/test_labels/test{i+1}_labels.json",
                path_dir + "/EndoVis2015/Tracking_Robotic_Testing/labels",
            )
        for i in range(4):
            covert_video_to_frames(
                path_dir + f"/EndoVis2015/Tracking_Robotic_Training/Dataset{i+1}/Video.avi",
                path_dir + f"/EndoVis2015/Tracking_Robotic_Training/Dataset{i+1}/raw",
                zero_fill=6,
            )
        for i in range(6):
            covert_video_to_frames(
                path_dir + f"/EndoVis2015/Tracking_Robotic_Testing/Dataset{i+1}/Video.avi",
                path_dir + f"/EndoVis2015/Tracking_Robotic_Testing/Dataset{i+1}/raw",
                zero_fill=4,
            )

    train_landmarks = {}
    for train_id in range(1, 5):
        json_labels = json.load(
            open(
                path_dir
                + f"/EndoVis2015/Tracking_Robotic_Training/labels/train{train_id}_labels.json"
            )
        )
        for annotation in json_labels:
            if annotation["annotations"] == []:
                continue
            landmarks = {}
            for name in class_names:
                landmarks[name] = torch.zeros((2, 2)).fill_(torch.nan)
                for labels in annotation["annotations"]:
                    if labels["class"] == name:
                        if landmarks[name][0, 0].isnan():
                            landmarks[name][0] = torch.tensor([labels["y"], labels["x"]])
                        else:
                            landmarks[name][1] = torch.tensor([labels["y"], labels["x"]])
            train_landmarks[
                path_dir
                + f"/EndoVis2015/Tracking_Robotic_Training/Dataset{train_id}/raw/"
                + os.path.basename(annotation["filename"].replace("_raw", ""))
            ] = landmarks

    train_landmarks_torch = torch.zeros((len(train_landmarks), len(class_names), 2, 2))
    for i, (key, value) in enumerate(train_landmarks.items()):
        for j, class_name in enumerate(class_names):
            train_landmarks_torch[i, j] = value[class_name]

    test_landmarks = {}
    for test_id in range(1, 7):
        json_labels = json.load(
            open(
                path_dir + f"/EndoVis2015/Tracking_Robotic_Testing/labels/test{test_id}_labels.json"
            )
        )
        for annotation in json_labels:
            if annotation["annotations"] == []:
                continue
            landmarks = {}
            for name in class_names:
                landmarks[name] = torch.zeros((2, 2)).fill_(torch.nan)
                for labels in annotation["annotations"]:
                    if labels["class"] == name:
                        if landmarks[name][0, 0].isnan():
                            landmarks[name][0] = torch.tensor([labels["y"], labels["x"]])
                        else:
                            landmarks[name][1] = torch.tensor([labels["y"], labels["x"]])
            test_landmarks[
                path_dir
                + f"/EndoVis2015/Tracking_Robotic_Testing/Dataset{test_id}/raw/"
                + os.path.basename(annotation["filename"])
            ] = landmarks

    test_landmarks_torch = torch.zeros((len(test_landmarks), len(class_names), 2, 2))
    for i, (key, value) in enumerate(test_landmarks.items()):
        for j, class_name in enumerate(class_names):
            test_landmarks_torch[i, j] = value[class_name]

    train_image_paths = list(train_landmarks.keys())
    test_image_paths = list(test_landmarks.keys())

    return train_image_paths, test_image_paths, train_landmarks_torch, test_landmarks_torch


def get_endovis2015_landmark_datasets(
    path_dir: str,
    class_names: list[str] = ["LeftClasperPoint", "RightClasperPoint"],
    train_transform=None,
    inference_transform=None,
    **kwargs,
) -> tuple[LandmarkDataset, LandmarkDataset]:
    (
        train_image_paths,
        test_image_paths,
        train_landmarks_torch,
        test_landmarks_torch,
    ) = get_endovis2015_dataset(path_dir, class_names)
    return LandmarkDataset(
        train_image_paths,
        train_landmarks_torch,
        class_names=class_names,
        transform=train_transform,
        **kwargs,
    ), LandmarkDataset(
        test_image_paths,
        test_landmarks_torch,
        class_names=class_names,
        transform=inference_transform,
        **kwargs,
    )


def get_endovis2015_heatmap_datasets(
    path_dir: str,
    class_names: list[str] = ["LeftClasperPoint", "RightClasperPoint"],
    sigma: float = 1.0,
    train_transform=None,
    inference_transform=None,
    **kwargs,
):
    (
        train_image_paths,
        test_image_paths,
        train_landmarks_torch,
        test_landmarks_torch,
    ) = get_endovis2015_dataset(path_dir, class_names)
    return (
        HeatmapDataset(
            train_image_paths,
            train_landmarks_torch,
            class_names=class_names,
            sigma=sigma,
            transform=train_transform,
            **kwargs,
        ),
        HeatmapDataset(
            test_image_paths,
            test_landmarks_torch,
            class_names=class_names,
            sigma=sigma,
            transform=inference_transform,
            **kwargs,
        ),
    )



"""Utility functions for the package."""

import glob
import json
import os
from typing import Optional

import cv2
import numpy as np
import torch
from tqdm import tqdm  # type: ignore


def get_paths(folder_path: str, extension: str) -> list[str]:
    """Retrieve all paths with a defined extension files recursively from the folder path.

    Args:
        folder_path (str): path to the folder
        extension (str): file extension

    Returns:
        list[str]: list of paths
    """
    # return all paths with a defined extension files recursively from the folder path
    return glob.glob(folder_path + "/**/*." + extension, recursive=True)


def annotation_to_landmark(json_obj: dict, class_names: list[str]) -> torch.Tensor:
    """
    Convert the annotation of LabelMe to landmarks.

    Args:
        json_obj (dict): annotation in json format
        class_names (list[str]): class names

    Returns:
        torch.Tensor: landmarks
    """
    shapes_list = json_obj["shapes"]
    landmarks = torch.ones((len(class_names), 2)) * torch.nan

    for i, name in enumerate(class_names):
        for shape in shapes_list:
            if shape["label"] == name:
                landmarks[i] = torch.Tensor(shape["points"][0][::-1])
                break
    return landmarks


def annotation_to_landmark_numpy(json_obj: dict, class_names: list[str]) -> np.ndarray:
    """
    Convert the annotation of LabelMe to landmarks (Numpy version).

    Args:
        json_obj (dict): annotation in json format
        class_names (list[str]): class names

    Returns:
        np.ndarray: landmarks
    """
    return annotation_to_landmark(json_obj, class_names).numpy()


def all_annotations_to_landmarks(paths: list[str], class_names: list[str]) -> torch.Tensor:
    """
    Convert all annotations of LabelMe to landmarks.

    Args:
        paths (list[str]): list of paths to the annotations
        class_names (list[str]): class names

    Returns:
        torch.Tensor: landmarks
    """
    landmarks = torch.zeros((len(paths), len(class_names), 2))
    for i, path in enumerate(paths):
        with open(path, encoding="utf8") as f:
            json_obj = json.load(f)
        landmarks[i] = annotation_to_landmark(json_obj, class_names)
    return landmarks


def all_annotations_to_landmarks_numpy(paths: list[str], class_names: list[str]) -> np.ndarray:
    """
    Convert all annotations of LabelMe to landmarks (Numpy version).

    Args:
        paths (list[str]): list of paths to the annotations
        class_names (list[str]): class names

    Returns:
        np.ndarray: landmarks
    """
    return all_annotations_to_landmarks(paths, class_names).numpy()


def get_angle(p1: torch.Tensor, p2: torch.Tensor, p3: torch.Tensor, radial=True) -> torch.Tensor:
    """
    Calculate the angle between three points. The angle is calculated in degrees or radians. The
    angle is calculated in the direction of p2 -> p3. If the angle is radial, the angle is between
    0 and 2pi. If the angle is not radial, the angle is between 0 and 360.

    Args:
        p1 (torch.Tensor): first point
        p2 (torch.Tensor): second point
        p3 (torch.Tensor): third point
        radial (bool, optional): whether the angle is radial. Defaults to True.

    Returns:
        torch.Tensor: angle
    """
    p1 = p1.reshape(-1, 2)
    p2 = p2.reshape(-1, 2)
    p3 = p3.reshape(-1, 2)
    angle = torch.acos(
        torch.sum((p2 - p1) * (p3 - p1), 1)
        / torch.sqrt(torch.sum((p2 - p1) ** 2, 1) * torch.sum((p3 - p1) ** 2, 1))
    )
    if radial:
        return angle % (2 * np.pi)
    return angle * 180 / np.pi % 360


def get_angle_numpy(p1: np.ndarray, p2: np.ndarray, p3: np.ndarray, radial=True) -> np.ndarray:
    """
    Calculate the angle between three points (Numpy version). See documentation of get_angle.

    Args:
        p1 (np.ndarray): first point
        p2 (np.ndarray): second point
        p3 (np.ndarray): third point
        radial (bool, optional): whether the angle is radial. Defaults to True.

    Returns:
        np.ndarray: angle
    """
    p1 = p1.reshape(-1, 2)
    p2 = p2.reshape(-1, 2)
    p3 = p3.reshape(-1, 2)
    angle = np.arccos(
        np.sum((p2 - p1) * (p3 - p1), 1)
        / np.sqrt(np.sum((p2 - p1) ** 2, 1) * np.sum((p3 - p1) ** 2, 1))
    )
    if radial:
        return angle % (2 * np.pi)
    return angle * 180 / np.pi % 360


def pixel_to_unit(
    landmarks: torch.Tensor,
    pixel_spacing: Optional[torch.Tensor] = None,
    dim: Optional[tuple[int, ...] | torch.Tensor] = None,
    dim_orig: Optional[torch.Tensor] = None,
    padding: Optional[torch.Tensor] = None,
):
    """Convert the landmarks from pixel to unit.

    Args:
        landmarks (torch.Tensor): landmarks
        pixel_spacing (Optional[torch.Tensor], optional): pixel spacing. Defaults to None.
        dim (Optional[tuple[int, ...] | torch.Tensor], optional): image size. Defaults to None.
        dim_orig (Optional[torch.Tensor], optional): original image size. Defaults to None.
        padding (Optional[torch.Tensor], optional): padding. Defaults to None.

    Returns:
        torch.Tensor: landmarks in units
    """
    spatial_dim = landmarks.shape[-1]
    if dim is not None:
        assert len(dim) == spatial_dim, f"dim must have {spatial_dim} elements."
    if dim_orig is not None:
        assert dim_orig.shape[-1] == spatial_dim, f"dim_orig must have {spatial_dim} elements."
    if len(landmarks.shape) == 4:
        added_dim: tuple[int, ...] = (1, 1)  # for multi-instance
    else:
        added_dim = (1,)
    if dim is not None and dim_orig is not None:
        if padding is None:
            padding = torch.zeros_like(dim_orig, device=landmarks.device)
        else:
            assert padding.shape == dim_orig.shape
        if pixel_spacing is None:
            pixel_spacing = torch.ones((1, spatial_dim), device=landmarks.device)
        dim = torch.as_tensor(dim, device=landmarks.device)
        landmarks_unresize = (
            landmarks
            * (dim_orig + 2.0 * padding).reshape((-1, *added_dim, spatial_dim))
            / dim.reshape((-1, *added_dim, spatial_dim))
        )
        landmarks_unpadded = landmarks_unresize - padding.reshape((-1, *added_dim, spatial_dim))
        return landmarks_unpadded * pixel_spacing.reshape((-1, *added_dim, spatial_dim))
    if dim is not None or dim_orig is not None:
        raise ValueError("dim and dim_orig must be both None or both not None.")
    if pixel_spacing is None:
        return landmarks
    return landmarks * pixel_spacing.reshape((-1, *added_dim, spatial_dim))


def pixel_to_unit_numpy(
    landmarks: np.ndarray,
    pixel_spacing: Optional[np.ndarray] = None,
    dim: Optional[tuple[int, ...] | np.ndarray] = None,
    dim_orig: Optional[np.ndarray] = None,
    padding: Optional[np.ndarray] = None,
) -> np.ndarray:
    # TODO: adjust to account for four dimensions (multi-instance)
    """
    Convert the landmarks from pixel to unit (Numpy version).

    Args:
        landmarks (np.ndarray): landmarks
        pixel_spacing (Optional[np.ndarray], optional): pixel spacing. Defaults to None.
        dim (Optional[tuple[int, ...] | np.ndarray], optional): image size. Defaults to None.
        dim_orig (Optional[np.ndarray], optional): original image size. Defaults to None.
        padding (Optional[np.ndarray], optional): padding. Defaults to None.

    Returns:
        np.ndarray: landmarks in units
    """
    spatial_dim = landmarks.shape[-1]
    if dim is not None:
        assert len(dim) == spatial_dim, f"dim must have {spatial_dim} elements."
    if dim_orig is not None:
        assert dim_orig.shape[-1] == spatial_dim, f"dim_orig must have {spatial_dim} elements."
    if len(landmarks.shape) == 4:
        added_dim: tuple[int, ...] = (1, 1)
    else:
        added_dim = (1,)
    if dim is not None and dim_orig is not None:
        if padding is None:
            padding = np.zeros_like(dim_orig)
        else:
            assert padding.shape == dim_orig.shape
        if pixel_spacing is None:
            pixel_spacing = np.ones((1, spatial_dim))
        dim = np.array(dim)
        landmarks_unresize = (
            landmarks
            * (dim_orig + 2.0 * padding).reshape((-1, *added_dim, spatial_dim))
            / dim.reshape((-1, *added_dim, spatial_dim))
        )
        landmarks_unpadded = landmarks_unresize - padding.reshape((-1, *added_dim, spatial_dim))
        return landmarks_unpadded * pixel_spacing.reshape((-1, *added_dim, spatial_dim))
    if dim is not None or dim_orig is not None:
        raise ValueError("dim and dim_orig must be both None or both not None.")
    if pixel_spacing is None:
        return landmarks
    return landmarks * pixel_spacing.reshape((-1, *added_dim, spatial_dim))


def covert_video_to_frames(video_path: str, frames_path: str, zero_fill: int = 6) -> None:
    """Convert a video to frames.

    Args:
        video_path (str): path to the video
        frames_path (str): path to the folder where the frames are saved
        zero_fill (int, optional): number of digits in the frame name. Defaults to 6.

    Returns:
        None
    """
    if not os.path.exists(frames_path):
        os.makedirs(frames_path)
    vidcap = cv2.VideoCapture(video_path)
    count = 0
    frame_count = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    tqdm.write(f"Transform video to frames: {frame_count} frames")
    for _ in tqdm(range(frame_count)):
        count += 1
        success, image = vidcap.read()
        if not success:
            print(f"Error in frame {count}")
            continue
        # save frame as PNG file
        cv2.imwrite(frames_path + f"/img_{str(count).zfill(zero_fill)}.png", image)



"""
Decoder module for retrieving coordinates and other statistics from heatmaps.
"""

from typing import Callable, List, Optional, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from scipy import optimize  # type: ignore

from landmarker.heatmap.generator import GaussianHeatmapGenerator


def coord_argmax(heatmap: torch.Tensor, spatial_dims: int = 2) -> torch.Tensor:
    """
    Returns the coordinates of the maximum value of the heatmap
    for each batch and channel (landmark).

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2) or (B, C, 3)
    """
    # TODO: Think about way to deal with ties (multiple maxima) in the output
    #    now we just take the first one
    if spatial_dims == 2:
        w = heatmap.shape[-1]
        idx_pred = torch.argmax(heatmap.flatten(-2, -1), dim=-1)
        idx_pred = torch.stack((idx_pred // w, idx_pred % w), dim=-1)
    elif spatial_dims == 3:
        d, h, w = heatmap.shape[-3:]
        idx_pred = torch.argmax(heatmap.flatten(-3, -1), dim=-1)
        idx_pred = torch.stack(
            (idx_pred // (h * w), (idx_pred % (h * w)) // w, idx_pred % w), dim=-1
        )
    else:
        raise ValueError(f"Spatial dimensions must be 2 or 3: {spatial_dims}")
    return idx_pred.int()


def coord_local_soft_argmax(
    heatmap: torch.Tensor, window: int = 5, t: float = 10.0, spatial_dims: int = 2
) -> torch.Tensor:
    """
    Returns coordiantes through applying the local soft-argmax function on the heatmaps.
        Source: "Subpixel Heatmap Regression for Facial Landmark Localization" - Bulat et al. (2021)

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        window (int): local window size
        t (float ): temperature that controls the resulting probability map
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2)
    """
    return coord_local_weighted(
        heatmap, window=window, spatial_dims=spatial_dims, activation="softmax", t=t
    )


def coord_local_weighted(
    heatmap: torch.Tensor,
    window: int = 9,
    spatial_dims: int = 2,
    activation: Optional[str] = None,
    t: float = 1.0,
) -> torch.Tensor:
    """
    Returns coordiantes through applying the local soft-argmax function on the heatmaps.
        Source: "Subpixel Heatmap Regression for Facial Landmark Localization" - Bulat et al. (2021)

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        window (int): local window size
        t (float ): temperature that controls the resulting probability map
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2)
    """
    if spatial_dims == 2:
        padded_heatmap = F.pad(heatmap, (window // 2, window // 2, window // 2, window // 2))
        b, c, h, w = padded_heatmap.shape

        argmax_coords = coord_argmax(heatmap) + window // 2

        mask = torch.zeros((b, c, h, w))
        for b1 in range(b):
            for c1 in range(c):
                for i in range(-(window // 2), (window // 2) + 1):
                    for j in range(-(window // 2), (window // 2) + 1):
                        mask[b1, c1, argmax_coords[b1, c1, 0] + i, argmax_coords[b1, c1, 1] + j] = 1
        masked_output = padded_heatmap[mask > 0].view(b, c, window, window)
        local_coords = coord_weighted_spatial_mean(
            masked_output, spatial_dims=spatial_dims, activation=activation, t=t
        )
        return local_coords + (argmax_coords - (window // 2) * 2)
    elif spatial_dims == 3:
        padding = [window // 2] * (2 * spatial_dims)  # Pad all spatial dimensions
        padded_heatmap = F.pad(heatmap, padding)
        b, c, d, h, w = padded_heatmap.shape

        argmax_coords = coord_argmax(heatmap, spatial_dims=3) + window // 2

        mask = torch.zeros((b, c, d, h, w))
        for b1 in range(b):
            for c1 in range(c):
                for z in range(-(window // 2), (window // 2) + 1):
                    for i in range(-(window // 2), (window // 2) + 1):
                        for j in range(-(window // 2), (window // 2) + 1):
                            mask[
                                b1,
                                c1,
                                argmax_coords[b1, c1, 0] + z,
                                argmax_coords[b1, c1, 1] + i,
                                argmax_coords[b1, c1, 2] + j,
                            ] = 1
        masked_output = padded_heatmap[mask > 0].view(b, c, window, window, window)

        local_coords = coord_weighted_spatial_mean(
            masked_output, spatial_dims=spatial_dims, activation=activation, t=t
        )
        return local_coords + (argmax_coords - (window // 2) * 2)
    else:
        raise ValueError(f"Spatial dimensions must be 2 or 3: {spatial_dims}")


def _activate_norm_heatmap(
    heatmap: torch.Tensor,
    spatial_dims: int = 2,
    activation: Optional[str] = "softmax",
    t: float = 1.0,
) -> torch.Tensor:
    dim = (-2, -1) if spatial_dims == 2 else (-3, -2, -1)
    if activation is not None:
        if activation == "softmax":
            heatmap = torch.exp(t * heatmap)
        elif activation == "sigmoid":
            heatmap = torch.sigmoid(heatmap)
        elif activation == "ReLU":
            heatmap = F.relu(heatmap)
        else:
            raise ValueError(f"Activation function {activation} not implemented.")
    return heatmap / torch.sum(heatmap, dim=dim, keepdim=True)


def coord_weighted_spatial_mean(
    heatmap: torch.Tensor,
    spatial_dims: int = 2,
    activation: Optional[str] = None,
    require_grad: bool = False,
    t: float = 1.0,
) -> torch.Tensor:
    """
    Returns the spatial weighted mean of the heatmap.
    Source: "UGLLI Face Alignment: Estimating Uncertainty with
        Gaussian Log-Likelihood Loss" - Kumar et al. (2019)

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        spatial_dims (int): number of spatial dimensions (2 or 3)
        activation (str): activation function to apply to the heatmap
        require_grad (bool): whether to require gradient for the coordinates

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2) or (B, C, 3)
    """
    heatmap = _activate_norm_heatmap(heatmap, spatial_dims=spatial_dims, activation=activation, t=t)
    if spatial_dims == 2:
        b, c, h, w = heatmap.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        xy = (
            torch.stack(torch.meshgrid(xs, ys, indexing="xy"), dim=2)
            .view(1, 1, h, w, 2)
            .repeat(b, c, 1, 1, 1)
        )
        return torch.sum(heatmap.unsqueeze(4) * xy, dim=(2, 3)).flip(-1)
    elif spatial_dims == 3:
        b, c, d, h, w = heatmap.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        zs = torch.arange(0, d, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        xyz = (
            torch.stack(torch.meshgrid(zs, ys, xs, indexing="ij"), dim=3)
            .view(1, 1, d, h, w, 3)
            .repeat(b, c, 1, 1, 1, 1)
        )
        return torch.sum(heatmap.unsqueeze(5) * xyz, dim=(2, 3, 4))
    else:
        raise ValueError(f"Spatial dimensions must be 2 or 3: {spatial_dims}")


def coord_soft_argmax_2d(
    heatmap: torch.Tensor, logit_scale: bool = False, require_grad: bool = False
) -> torch.Tensor:
    """
    Returns the spatial mean over the softmax distribution of the heatmap,
    but for a 2D heatmap, without the batch and channel dimensions.

    Args:
        heatmap (torch.Tensor): heatmap of shape (H, W)
        logit_scale (bool): whether the input is logit scaled
        require_grad (bool): whether to require gradient for the coordinates
    """
    activation = "softmax" if not logit_scale else None
    return coord_weighted_spatial_mean(
        heatmap.unsqueeze(0).unsqueeze(0),
        spatial_dims=2,
        activation=activation,
        require_grad=require_grad,
    )[0, 0]


def heatmap_to_coord(
    heatmap: torch.Tensor,
    offset_coords: int = 0,
    method: str = "argmax",
    spatial_dims: int = 2,
    require_grad: bool = False,
) -> torch.Tensor:
    """
    Returns the retrieved coordinates via specified method from a heatmap. The offset_coords is used
    to remove the first offset_coords coordinates from the heatmap. This is used to remove the
    background class (if present).

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        offset_coords (int): number of coordinates to remove
        method (str): method to retrieve the coordinates
        spatial_dims (int): number of spatial dimensions (2 or 3)
        require_grad (bool): whether to require gradient for the coordinates

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2) or (B, C, 3)
    """
    heatmap = heatmap[:, offset_coords:]
    if method == "argmax":
        return coord_argmax(heatmap, spatial_dims=spatial_dims)
    if method == "local_soft_argmax":
        return coord_local_soft_argmax(heatmap, spatial_dims=spatial_dims)
    if method == "weighted_spatial_mean_relu":
        return coord_weighted_spatial_mean(
            heatmap, spatial_dims=spatial_dims, activation="ReLU", require_grad=require_grad
        )
    if method == "soft_argmax":
        return coord_weighted_spatial_mean(
            heatmap, spatial_dims=spatial_dims, activation="softmax", require_grad=require_grad
        )
    if method == "weighted_spatial_mean":
        return coord_weighted_spatial_mean(
            heatmap, spatial_dims=spatial_dims, require_grad=require_grad
        )
    if method == "weighted_spatial_mean_sigmoid":
        return coord_weighted_spatial_mean(
            heatmap, spatial_dims=spatial_dims, activation="sigmoid", require_grad=require_grad
        )
    raise ValueError("Method not implemented.")


def heatmap_to_coord_enlarge(
    heatmap: torch.Tensor,
    offset_coords: int = 0,
    method: str = "argmax",
    enlarge_factor: int = 1,
    enlarge_dim: Optional[tuple[int, ...]] = None,
    enlarge_mode: str = "bilinear",
    spatial_dims: int = 2,
) -> torch.Tensor:
    """
    Returns the retrieved coordinates via specified method from an enlarged heatmap.
    The offset_coords is used to remove the first offset_coords coordinates from the heatmap.
    This is used to remove the background class (if present).

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        offset_coords (int): number of coordinates to remove
        method (str): method to retrieve the coordinates
        enlarge_factor (int): factor to enlarge the heatmap
        enlarge_dim (tuple[int, ...] or None): dimensions to enlarge the heatmap to
        enlarge_mode (str): interpolation mode to enlarge the heatmap
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2)
    """
    heatmap = heatmap[:, offset_coords:]
    if spatial_dims == 2:
        if enlarge_dim is not None:
            assert len(enlarge_dim) == 2, "enlarge_dim must be a tuple of 2 elements"
            heatmap_scaled = F.interpolate(
                heatmap, size=enlarge_dim, mode=enlarge_mode, align_corners=False
            )
        else:
            heatmap_scaled = F.interpolate(
                heatmap, scale_factor=enlarge_factor, mode=enlarge_mode, align_corners=False
            )
        coords = heatmap_to_coord(heatmap_scaled, method=method)
        if enlarge_dim is not None:
            coords = coords / (
                torch.tensor(enlarge_dim, dtype=torch.float).view((1, 1, 2)).to(heatmap.device)
                / torch.tensor(heatmap.shape[-2:], dtype=torch.float)
                .view((1, 1, 2))
                .to(heatmap.device)
            )
        else:
            coords = coords / enlarge_factor
        return coords
    elif spatial_dims == 3:
        if enlarge_dim is not None:
            assert len(enlarge_dim) == 3, "enlarge_dim must be a tuple of 3 elements"
            heatmap_scaled = F.interpolate(
                heatmap, size=enlarge_dim, mode=enlarge_mode, align_corners=False
            )
        else:
            heatmap_scaled = F.interpolate(
                heatmap, scale_factor=enlarge_factor, mode=enlarge_mode, align_corners=False
            )
        coords = heatmap_to_coord(heatmap_scaled, method=method, spatial_dims=3)
        if enlarge_dim is not None:
            coords = coords / (
                torch.tensor(enlarge_dim, dtype=torch.float).view((1, 1, 3)).to(heatmap.device)
                / torch.tensor(heatmap.shape[-3:], dtype=torch.float)
                .view((1, 1, 3))
                .to(heatmap.device)
            )
        else:
            coords = coords / enlarge_factor
        return coords
    else:
        raise ValueError(f"Spatial dimensions must be 2 or 3: {spatial_dims}")


def coord_weighted_spatial_mean_cov(
    heatmap: torch.Tensor,
    spatial_dims: int = 2,
    require_grad: bool = False,
    activation: Optional[str] = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Returns the spatial weighted mean and the weighted sample covariance of the possitive elements
    of the heatmap by the heatmap values.
        source: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_covariance

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        spatial_dims (int): number of spatial dimensions (2 or 3)
        require_grad (bool): whether to require gradient for the coordinates
        activation (str): activation function to apply to the heatmap

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2) or (B, C, 3)
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2) or (B, C, 3, 3)
    """
    heatmap = _activate_norm_heatmap(heatmap, spatial_dims=spatial_dims, activation=activation)
    if spatial_dims == 2:
        b, c, h, w = heatmap.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        xs, ys = torch.meshgrid(xs, ys, indexing="xy")
        yx = torch.stack((ys, xs), dim=2).view(1, 1, h, w, 2).repeat(b, c, 1, 1, 1)
        mean_coords = torch.sum(heatmap.unsqueeze(4) * yx, dim=(2, 3))
        dist_x = xs - mean_coords[:, :, 1].unsqueeze(-1).unsqueeze(-1)
        dist_y = ys - mean_coords[:, :, 0].unsqueeze(-1).unsqueeze(-1)
        dist = torch.stack(
            (dist_y.flatten(start_dim=-2, end_dim=-1), dist_x.flatten(start_dim=-2, end_dim=-1)),
            dim=-1,
        )  # (B, C, H*W, 2)

        covariances = (
            (heatmap.flatten(start_dim=-2, end_dim=-1).unsqueeze(-1) * dist).transpose(-1, -2)
            @ dist
        ).view(b, c, 2, 2)

        v2 = torch.sum(heatmap**2, dim=(2, 3)).unsqueeze(-1).unsqueeze(-1)
        return mean_coords, covariances / (1 - v2)
    elif spatial_dims == 3:
        b, c, d, h, w = heatmap.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        zs = torch.arange(0, d, dtype=torch.float32, requires_grad=require_grad).to(heatmap.device)
        zs, ys, xs = torch.meshgrid(zs, ys, xs, indexing="ij")
        zyx = torch.stack((zs, ys, xs), dim=3).view(1, 1, d, h, w, 3).repeat(b, c, 1, 1, 1, 1)

        mean_coords = torch.sum(heatmap.unsqueeze(5) * zyx, dim=(2, 3, 4))

        dist_x = xs - mean_coords[:, :, 2].unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        dist_y = ys - mean_coords[:, :, 1].unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        dist_z = zs - mean_coords[:, :, 0].unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        dist = torch.stack(
            (
                dist_z.flatten(start_dim=-3, end_dim=-1),
                dist_y.flatten(start_dim=-3, end_dim=-1),
                dist_x.flatten(start_dim=-3, end_dim=-1),
            ),
            dim=-1,
        )

        covariances = (
            (heatmap.flatten(start_dim=-3, end_dim=-1).unsqueeze(-1) * dist).transpose(-1, -2)
            @ dist
        ).view(b, c, 3, 3)

        v2 = torch.sum(heatmap**2, dim=(2, 3, 4)).unsqueeze(-1).unsqueeze(-1)

        return mean_coords, covariances / (1 - v2)
    else:
        raise ValueError(f"Spatial dimensions must be 2 or 3: {spatial_dims}")


def heatmap_to_coord_cov(
    heatmap: torch.Tensor,
    method: str = "soft_argmax",
    require_grad: bool = True,
    spatial_dims: int = 2,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Returns the modal coordinates and covariance matrix from a heatmap.

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        method (str): method to retrieve the coordinates
        require_grad (bool): whether to require gradient for the coordinates
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2) or (B, C, 3)
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2) or (B, C, 3, 3)
    """
    if method == "soft_argmax":
        return coord_weighted_spatial_mean_cov(
            heatmap, spatial_dims=spatial_dims, require_grad=require_grad, activation="softmax"
        )
    if method == "weighted_spatial_mean":
        return coord_weighted_spatial_mean_cov(
            heatmap, spatial_dims=spatial_dims, require_grad=require_grad
        )
    if method == "weighted_spatial_mean_sigmoid":
        return coord_weighted_spatial_mean_cov(
            heatmap, spatial_dims=spatial_dims, require_grad=require_grad, activation="sigmoid"
        )
    if method == "weighted_spatial_mean_relu":
        return coord_weighted_spatial_mean_cov(
            heatmap, spatial_dims=spatial_dims, require_grad=require_grad, activation="ReLU"
        )
    raise ValueError("Method not implemented.")


def coord_cov_from_gaussian_ls(
    heatmap: torch.Tensor,
    gamma: Optional[float] = None,
    ls_library: str = "scipy",
    spatial_dims: int = 2,
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Returns the modal coordinates and covariance matrix from a heatmap through fitting the heatmap
    on Gaussian distribution with a specicic scaling factor gamma with help of least squares
    optimization.

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W)
        gamma (float): gamma parameter of the gaussian heatmap generator
        ls_library (str): library to use for least squares optimization. (scipy or pytorch)
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2)
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2)
    """
    # TODO: see if we can use a pytorch implementation (e.g, pytorch-minimize seems to be broken)
    if spatial_dims != 2:
        raise ValueError(f"Spatial dimensions must be 2: {spatial_dims}")
    if ls_library == "scipy":
        return coord_cov_from_gaussian_ls_scipy(heatmap, gamma=gamma)
    raise ValueError("Method not implemented.")


def coord_cov_from_gaussian_ls_scipy(
    heatmap: torch.Tensor, gamma: Optional[float] = None, spatial_dims: int = 2
) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Returns the modal coordinates and covariance matrix from a heatmap through fitting the heatmap
    on Gaussian distribution with a specicic scaling factor gamma with help of least squares
    optimization.

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W)
        gamma (float): gamma parameter of the gaussian heatmap generator
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): coordinates of shape (B, C, 2)
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2)
    """

    def generate_gaussian(heatmap: torch.Tensor) -> Callable:
        """
        Returns the gaussian function for the given landmarks, sigma and rotation.
        """
        gaussian_generator = GaussianHeatmapGenerator(
            1,
            gamma=gamma,
            heatmap_size=(heatmap.shape[-2], heatmap.shape[-1]),
            learnable=False,
        ).to(heatmap.device)

        def fun_to_minimize(x):
            gaussian_generator.set_sigmas(x[:2])
            gaussian_generator.set_rotation(x[2])
            return (
                (
                    heatmap
                    - gaussian_generator(torch.Tensor(x[3:]).view((1, 1, 2))).view(heatmap.shape)
                )
                .flatten()
                .detach()
                .cpu()
                .numpy()
            )

        return fun_to_minimize

    if spatial_dims != 2:
        raise ValueError(f"Spatial dimensions must be 2: {spatial_dims}")
    b, c, _, _ = heatmap.shape
    coords = torch.zeros((b, c, 2)).to(heatmap.device)
    covs = torch.zeros((b, c, 2, 2)).to(heatmap.device)
    for b1 in range(b):
        for c1 in range(c):
            init_coord = coord_argmax(heatmap[b1, c1].unsqueeze(0).unsqueeze(0))[0, 0]
            result = optimize.least_squares(
                generate_gaussian(heatmap[b1, c1]),
                np.array([1, 1, 0, init_coord[0].item(), init_coord[1].item()]),
                method="trf",
            )
            x = result.x
            coords[b1, c1] = torch.tensor([x[3], x[4]], dtype=torch.float)
            rotation = torch.tensor(
                [[np.cos(x[2]), -np.sin(x[2])], [np.sin(x[2]), np.cos(x[2])]], dtype=torch.float
            )
            diagonal = torch.diag(torch.tensor(x[:2] ** 2, dtype=torch.float))
            covs[b1, c1] = torch.mm(torch.mm(rotation, diagonal), rotation.t())
    return coords, covs


def cov_from_gaussian_ls(
    heatmap: torch.Tensor,
    coords: torch.Tensor,
    gamma: Optional[float] = None,
    ls_library: str = "scipy",
    spatial_dims: int = 2,
) -> torch.Tensor:
    """
    Returns covariance matrix from a heatmap through fitting the heatmap on Gaussian distribution
    with a specicic scaling factor gamma and specified coordinates
    with help of least squares optimization.

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W)
        coords (torch.Tensor): coordinates of shape (B, C, 2)
        gamma (float): gamma parameter of the gaussian heatmap generator
        ls_library (str): library to use for least squares optimization. (scipy or pytorch)

    Returns:
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2)
    """
    # TODO: see if we can use a pytorch implementation (e.g, pytorch-minimize seems to be broken)
    if spatial_dims != 2:
        raise ValueError(f"Spatial dimensions must be 2: {spatial_dims}")
    if ls_library == "scipy":
        return cov_from_gaussian_ls_scipy(heatmap, coords, gamma=gamma)
    raise ValueError("Method not implemented.")


def cov_from_gaussian_ls_scipy(
    heatmap: torch.Tensor,
    coords: torch.Tensor,
    gamma: Optional[float] = None,
    spatial_dims: int = 2,
) -> torch.Tensor:
    """
    Returns the covariance matrix from a heatmap through fitting the heatmap on Gaussian
    distribution with a specicic scaling factor gamma and retrieved coordinates with specified
    decoding method with help of least squares optimization.

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W)
        coords (torch.Tensor): coordinates of shape (B, C, 2)
        gamma (float): gamma parameter of the gaussian heatmap generator
        spatial_dims (int): number of spatial dimensions (2 or 3)

    Returns:
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2)
    """

    def generate_gaussian(landmarks: torch.Tensor, heatmap: torch.Tensor) -> Callable:
        """
        Returns the gaussian function for the given landmarks, sigma and rotation.
        """
        gaussian_generator = GaussianHeatmapGenerator(
            1,
            gamma=gamma,
            heatmap_size=(heatmap.shape[-2], heatmap.shape[-1]),
            learnable=False,
        ).to(heatmap.device)

        def fun_to_minimize(x):
            gaussian_generator.set_sigmas(x[:2])
            gaussian_generator.set_rotation(x[2])
            return (
                (heatmap - gaussian_generator(landmarks.view(1, 1, 2)).view(heatmap.shape))
                .flatten()
                .detach()
                .cpu()
                .numpy()
            )

        return fun_to_minimize

    if spatial_dims != 2:
        raise ValueError(f"Spatial dimensions must be 2: {spatial_dims}")
    covs = torch.zeros((heatmap.shape[0], heatmap.shape[1], 2, 2)).to(heatmap.device)
    for b in range(heatmap.shape[0]):
        for c in range(heatmap.shape[1]):
            result = optimize.least_squares(
                generate_gaussian(coords[b, c], heatmap[b, c]), np.array([1, 1, 0]), method="trf"
            )
            x = result.x
            rotation = torch.tensor(
                [[np.cos(x[2]), -np.sin(x[2])], [np.sin(x[2]), np.cos(x[2])]], dtype=torch.float
            )
            diagonal = torch.diag(torch.tensor(x[:2] ** 2, dtype=torch.float))
            covs[b, c] = torch.mm(torch.mm(rotation, diagonal), rotation.t())
    return covs


def weighted_sample_cov(
    heatmap: torch.Tensor,
    coords: torch.Tensor,
    spatial_dims: int = 2,
    activation: Optional[str] = None,
) -> torch.Tensor:
    """
    Calculate the covariance matrix from a heatmap by calculating the mean of the
    heatmap values weighted by the heatmap values.
    source: https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_covariance

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W) or (B, C, D, H, W)
        coords (torch.Tensor): coordinates of shape (B, C, 2) or (B, C, 3)
        spatial_dims (int): number of spatial dimensions (2 or 3)
        activation (str): activation function to apply to the heatmap

    Returns:
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2) or (B, C, 3, 3)
    """
    heatmap = _activate_norm_heatmap(heatmap, spatial_dims=spatial_dims, activation=activation)
    if spatial_dims == 2:
        assert coords.shape[-1] == 2, f"Coordinates must have 2 elements: {coords.shape[-1]}"
        b, c, h, w = heatmap.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=False).to(heatmap.device)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=False).to(heatmap.device)
        xs, ys = torch.meshgrid(xs, ys, indexing="xy")

        dist_x = xs - coords[:, :, 1].unsqueeze(-1).unsqueeze(-1)
        dist_y = ys - coords[:, :, 0].unsqueeze(-1).unsqueeze(-1)
        dist = torch.stack(
            (dist_y.flatten(start_dim=-2, end_dim=-1), dist_x.flatten(start_dim=-2, end_dim=-1)),
            dim=-1,
        )  # (B, C, H*W, 2)

        covariances = (
            (heatmap.flatten(start_dim=-2, end_dim=-1).unsqueeze(-1) * dist).transpose(-1, -2)
            @ dist
        ).view(b, c, 2, 2)

        v2 = torch.sum(heatmap**2, dim=(2, 3)).unsqueeze(-1).unsqueeze(-1)
        assert covariances.shape == (b, c, 2, 2)
        return covariances / (1 - v2)
    elif spatial_dims == 3:
        assert coords.shape[-1] == 3, f"Coordinates must have 3 elements: {coords.shape[-1]}"
        b, c, d, h, w = heatmap.shape
        xs = torch.arange(0, w, dtype=torch.float32, requires_grad=False).to(heatmap.device)
        ys = torch.arange(0, h, dtype=torch.float32, requires_grad=False).to(heatmap.device)
        zs = torch.arange(0, d, dtype=torch.float32, requires_grad=False).to(heatmap.device)
        zs, ys, xs = torch.meshgrid(zs, ys, xs, indexing="ij")

        dist_x = xs - coords[:, :, 2].unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        dist_y = ys - coords[:, :, 1].unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        dist_z = zs - coords[:, :, 0].unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        dist = torch.stack(
            (
                dist_z.flatten(start_dim=-3, end_dim=-1),
                dist_y.flatten(start_dim=-3, end_dim=-1),
                dist_x.flatten(start_dim=-3, end_dim=-1),
            ),
            dim=-1,
        )

        covariances = (
            (heatmap.flatten(start_dim=-3, end_dim=-1).unsqueeze(-1) * dist).transpose(-1, -2)
            @ dist
        ).view(b, c, 3, 3)

        v2 = torch.sum(heatmap**2, dim=(2, 3, 4)).unsqueeze(-1).unsqueeze(-1)
        assert covariances.shape == (b, c, 3, 3)
        return covariances / (1 - v2)
    else:
        raise ValueError(f"Spatial dimensions must be 2 or 3: {spatial_dims}")


def windowed_weigthed_sample_cov(
    heatmap: torch.Tensor,
    coords: torch.Tensor,
    spatial_dims: int = 2,
    activation: Optional[str] = None,
) -> torch.Tensor:
    """
    Calculate the covariance matrix from a heatmap by calculating the mean of the
    heatmap values weighted by the heatmap values. The window is determined by the
    distance to the closest edge.

    Args:
        heatmap (torch.Tensor): heatmap of shape (B, C, H, W)
        coords (torch.Tensor): coordinates of shape (B, C, 2)
        spatial_dims (int): number of spatial dimensions (2 or 3)
        activation (str): activation function to apply to the heatmap

    Returns:
        (torch.Tensor): covariance matrix of shape (B, C, 2, 2)
    """
    if spatial_dims != 2:
        raise ValueError(f"Spatial dimensions must be 2: {spatial_dims}")
    b, c, h, w = heatmap.shape
    covs = torch.zeros((b, c, 2, 2)).to(heatmap.device)
    for b in range(b):
        for c in range(c):
            window = int(
                min(h - coords[b, c, 0], coords[b, c, 0], w - coords[b, c, 1], coords[b, c, 1])
            )
            covs[b, c] = weighted_sample_cov(
                heatmap[
                    b,
                    c,
                    coords[b, c, 0].int() - window : coords[b, c, 0].int() + window + 1,
                    coords[b, c, 1].int() - window : coords[b, c, 1].int() + window + 1,
                ]
                .unsqueeze(0)
                .unsqueeze(0),
                torch.tensor([[[window, window]]], dtype=torch.float).to(heatmap.device),
                spatial_dims=spatial_dims,
                activation=activation,
            )
    return covs


def heatmap_to_multiple_coord(
    heatmaps: torch.Tensor,
    window: int = 5,
    threshold: Optional[float] = None,
    method: str = "argmax",
    spatial_dims: int = 2,
) -> Tuple[List[List[List[torch.Tensor]]], List[List[List[float]]]]:

    # Original input shape handling
    if spatial_dims != 2:
        raise ValueError(f"Spatial dimensions must be 2: {spatial_dims}")

    # Dimension handling identical to original
    orig_dim = heatmaps.dim()
    if orig_dim == 2:
        heatmaps = heatmaps.unsqueeze(0).unsqueeze(0)
    elif orig_dim == 3:
        heatmaps = heatmaps.unsqueeze(0)
    B, C, H, W = heatmaps.shape

    # Main processing
    if method == "argmax":
        coords, scores = non_maximum_surpression(heatmaps, window)
    else:
        raise ValueError("Method not implemented.")

    # Post-processing identical to original
    return process_multiple_coord_results(coords, scores, threshold, B, C, orig_dim)


def process_multiple_coord_results(coords, scores, threshold, B, C, orig_dim):
    # Identical sorting and threshold logic to original
    out = []
    score_out = []

    for b in range(B):
        batch_coords = []
        batch_scores = []
        for c in range(C):
            # Sort descending by score
            sc = scores[b][c]
            idx = sorted(range(len(sc)), key=lambda k: sc[k], reverse=True)
            sorted_coords = [coords[b][c][i] for i in idx]
            sorted_scores = [sc[i] for i in idx]

            # Apply threshold
            if threshold is not None:
                cutoff = next(
                    (i for i, s in enumerate(sorted_scores) if s < threshold), len(sorted_scores)
                )
                sorted_coords = sorted_coords[:cutoff]
                sorted_scores = sorted_scores[:cutoff]

            batch_coords.append(sorted_coords)
            batch_scores.append(sorted_scores)

        out.append(batch_coords)
        score_out.append(batch_scores)

    # Match original output shapes
    if orig_dim == 2:
        return out[0][0], score_out[0][0]
    elif orig_dim == 3:
        return out[0], score_out[0]
    return out, score_out


def non_maximum_surpression(
    heatmaps: torch.Tensor, window: int
) -> Tuple[List[List[List[torch.Tensor]]], List[List[List[float]]]]:

    B, C, H, W = heatmaps.shape
    device = heatmaps.device
    pad = window // 2
    padded = F.pad(heatmaps, (pad, pad, pad, pad), mode="constant", value=-float("inf"))

    # Step 1: Find initial candidates using original's block strategy
    grid_y = torch.arange(0, H, window, device=device)
    grid_x = torch.arange(0, W, window, device=device)
    gy, gx = torch.meshgrid(grid_y, grid_x, indexing="ij")
    candidates = torch.stack([gy.flatten(), gx.flatten()], dim=1)  # [K, 2]

    # Step 2: Find local maxima within each block
    max_vals = torch.zeros(B, C, len(candidates), device=device)
    max_pos = torch.zeros(B, C, len(candidates), 2, dtype=torch.long, device=device)

    for i, (y, x) in enumerate(candidates):
        block = padded[:, :, y : y + window, x : x + window]
        vals, idx = block.flatten(2).max(dim=-1)  # [B, C]
        max_vals[:, :, i] = vals
        max_pos[:, :, i] = torch.stack([y + idx // window, x + idx % window], dim=-1)

    # Step 3: Verify maxima in their neighborhoods (original's 2nd check)
    valid = torch.zeros_like(max_vals, dtype=torch.bool)
    for i in range(len(candidates)):
        y, x = max_pos[:, :, i, 0], max_pos[:, :, i, 1]

        # Get neighborhood bounds (matching original's clamp logic)
        y0 = torch.clamp(y - pad, 0)
        y1 = torch.clamp(y + pad + 1, max=padded.shape[2])
        x0 = torch.clamp(x - pad, 0)
        x1 = torch.clamp(x + pad + 1, max=padded.shape[3])

        # Check if center is max in neighborhood
        for b in range(B):
            for c in range(C):
                neighborhood = padded[b, c, y0[b, c] : y1[b, c], x0[b, c] : x1[b, c]]
                if neighborhood.numel() > 0:
                    valid[b, c, i] = max_vals[b, c, i] >= neighborhood.max()

    # Step 4: Collect valid coordinates and deduplicate
    coords = [[[] for _ in range(C)] for _ in range(B)]  # type: ignore
    scores = [[[] for _ in range(C)] for _ in range(B)]  # type: ignore

    for b in range(B):
        for c in range(C):
            mask = valid[b, c]
            pos = max_pos[b, c][mask]
            vals = max_vals[b, c][mask]

            # Deduplicate using unique coordinates
            unique_coords, inverse = torch.unique(pos, dim=0, return_inverse=True)
            unique_vals = torch.zeros_like(unique_coords[:, 0], dtype=vals.dtype)
            unique_vals.scatter_reduce_(0, inverse, vals, reduce="amax", include_self=False)

            # Convert to original image coordinates
            final_coords = unique_coords.float() - pad  # Remove padding offset
            coords[b][c] = [final_coords[i] for i in range(len(final_coords))]
            scores[b][c] = unique_vals.tolist()

    return coords, scores



"""Tests for the metrics module."""

import numpy as np
import torch

from src.landmarker.metrics.metrics import multi_instance_point_error, point_error, sdr


def test_point_error():
    """Test the point_error function."""
    true_landmarks = torch.Tensor([[[10, 20], [30, 40]], [[50, 60], [70, 80]]])
    pred_landmarks = torch.Tensor([[[10, 20], [30, 42]], [[52, 62], [75, 82]]])
    pixel_spacing = torch.Tensor([[0.1, 0.2], [0.3, 0.4]])
    dim = (100, 200)
    dim_orig = torch.Tensor([[50, 100], [150, 200]])

    expected_result = torch.Tensor(
        [
            [0, 2 * 0.2 * 100 / 200],
            [
                np.sqrt((2 * 0.3 * 150 / 100) ** 2 + (2 * 0.4 * 200 / 200) ** 2),
                np.sqrt((5 * 0.3 * 150 / 100) ** 2 + (2 * 0.4 * 200 / 200) ** 2),
            ],
        ]
    )

    result = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=pixel_spacing,
        reduction="none",
        padding=None,
    )

    assert torch.allclose(result, expected_result)

    expected_result = torch.mean(expected_result)
    result = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=pixel_spacing,
        reduction="mean",
    )

    assert torch.allclose(result, expected_result)

    expected_result = torch.Tensor(
        [
            [0, 2 * 100 / 200],
            [
                np.sqrt((2 * 150 / 100) ** 2 + (2 * 200 / 200) ** 2),
                np.sqrt((5 * 150 / 100) ** 2 + (2 * 200 / 200) ** 2),
            ],
        ]
    )

    result = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=None,
        reduction="none",
    )

    assert torch.allclose(result, expected_result)

    expected_result = torch.Tensor(
        [[0, 2], [np.sqrt((2) ** 2 + (2) ** 2), np.sqrt((5) ** 2 + (2) ** 2)]]
    )

    try:
        result = point_error(
            true_landmarks,
            pred_landmarks,
            dim=dim,
            dim_orig=dim_orig,
            pixel_spacing=None,
            reduction="not supported",
        )
        assert False
    except ValueError:
        assert True


def test_point_error_3d():
    """Test the point_error function. 3D case."""
    true_landmarks = torch.Tensor([[[10, 20, 34], [30, 40, 20]], [[50, 60, 14], [70, 80, 45]]])
    pred_landmarks = torch.Tensor([[[10, 20, 34], [30, 42, 20]], [[52, 62, 14], [75, 82, 38]]])
    pixel_spacing = torch.Tensor([[0.1, 0.2, 0.1], [0.3, 0.4, 0.2]])
    dim = (100, 200, 120)
    dim_orig = torch.Tensor([[50, 100, 200], [150, 200, 200]])

    expected_result = torch.Tensor(
        [
            [0, 2 * 0.2 * 100 / 200],
            [
                np.sqrt((2 * 0.3 * 150 / 100) ** 2 + (2 * 0.4 * 200 / 200) ** 2),
                np.sqrt(
                    (5 * 0.3 * 150 / 100) ** 2
                    + (2 * 0.4 * 200 / 200) ** 2
                    + (7 * 0.2 * 200 / 120) ** 2
                ),
            ],
        ]
    )

    result = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=pixel_spacing,
        reduction="none",
        padding=None,
    )

    assert torch.allclose(result, expected_result)

    expected_result = torch.mean(expected_result)
    result = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=pixel_spacing,
        reduction="mean",
    )

    assert torch.allclose(result, expected_result)

    expected_result = torch.Tensor(
        [
            [0, 2 * 100 / 200],
            [
                np.sqrt((2 * 150 / 100) ** 2 + (2 * 200 / 200) ** 2),
                np.sqrt((5 * 150 / 100) ** 2 + (2 * 200 / 200) ** 2 + (7 * 200 / 120) ** 2),
            ],
        ]
    )

    result = point_error(
        true_landmarks,
        pred_landmarks,
        dim=dim,
        dim_orig=dim_orig,
        pixel_spacing=None,
        reduction="none",
    )

    assert torch.allclose(result, expected_result)

    try:
        result = point_error(
            true_landmarks,
            pred_landmarks,
            dim=dim,
            dim_orig=dim_orig,
            pixel_spacing=None,
            reduction="not supported",
        )
        assert False
    except ValueError:
        assert True


def test_sdr_single_radius():
    """
    Test the sdr function for a single radius.
    """
    # create some dummy itorchut tensors
    radius = 5
    true_landmarks = torch.Tensor(np.random.rand(4, 10, 2))
    pred_landmarks = torch.Tensor(np.random.rand(4, 10, 2))

    # calculate the SDR
    sdr_ = sdr(radius, true_landmarks, pred_landmarks)

    # check that the output is a scalar value
    assert isinstance(sdr_, float)

    # check that the output is between 0 and 100
    assert 0 <= sdr_ <= 100


def test_sdr_single_radius_3d():
    """
    Test the sdr function for a single radius. 3D case.
    """
    # create some dummy itorchut tensors
    radius = 5
    true_landmarks = torch.Tensor(np.random.rand(4, 10, 3))
    pred_landmarks = torch.Tensor(np.random.rand(4, 10, 3))

    # calculate the SDR
    sdr_ = sdr(radius, true_landmarks, pred_landmarks)

    # check that the output is a scalar value
    assert isinstance(sdr_, float)

    # check that the output is between 0 and 100
    assert 0 <= sdr_ <= 100


def test_sdr_multiple_radii():
    """
    Test the sdr function for multiple radii.
    """
    # create some dummy itorchut tensors
    radii = [1, 5, 10]
    true_landmarks = torch.Tensor(np.random.rand(4, 10, 2))
    pred_landmarks = torch.Tensor(np.random.rand(4, 10, 2))

    # calculate the SDR for multiple radii
    sdr_dict = sdr(radii, true_landmarks, pred_landmarks)

    # check that the output is a dictionary
    assert isinstance(sdr_dict, dict)

    # check that the dictionary has the correct keys
    assert set(sdr_dict.keys()) == set(radii)

    # check that the dictionary values are between 0 and 100
    for sdr_ in sdr_dict.values():
        assert 0 <= sdr_ <= 100


def test_sdr_multiple_radii_3d():
    """
    Test the sdr function for multiple radii. 3D case.
    """
    # create some dummy itorchut tensors
    radii = [1, 5, 10]
    true_landmarks = torch.Tensor(np.random.rand(4, 10, 3))
    pred_landmarks = torch.Tensor(np.random.rand(4, 10, 3))

    # calculate the SDR for multiple radii
    sdr_dict = sdr(radii, true_landmarks, pred_landmarks)

    # check that the output is a dictionary
    assert isinstance(sdr_dict, dict)

    # check that the dictionary has the correct keys
    assert set(sdr_dict.keys()) == set(radii)

    # check that the dictionary values are between 0 and 100
    for sdr_ in sdr_dict.values():
        assert 0 <= sdr_ <= 100


def test_multi_instance_point_error():
    # Create lists of true and predicted landmarks
    true_landmarks = torch.Tensor([[[[1, 2], [3, 4], [5, 6]]]])
    pred_landmarks = [torch.Tensor([1, 2]), torch.Tensor([3, 4]), torch.Tensor([5, 6])]

    assert true_landmarks.shape == (1, 1, 3, 2)  # (batch, class, instance, dim)

    # Call the function
    pe, tp, fp, fn, _ = multi_instance_point_error(true_landmarks, pred_landmarks)

    # Check that the output has the expected values
    assert pe == 0
    assert tp == 3
    assert fp == 0


def test_multi_instance_point_error_3d():
    # Create lists of true and predicted landmarks
    true_landmarks = torch.Tensor([[[[1, 2, 2], [3, 4, 2], [5, 6, 5]]]])
    pred_landmarks = [torch.Tensor([1, 2, 2]), torch.Tensor([3, 4, 2]), torch.Tensor([5, 6, 5])]

    assert true_landmarks.shape == (1, 1, 3, 3)  # (batch, class, instance, dim)

    # Call the function
    pe, tp, fp, fn, _ = multi_instance_point_error(true_landmarks, pred_landmarks)

    # Check that the output has the expected values
    assert pe == 0
    assert tp == 3
    assert fp == 0


def test_multi_instance_point_error_different_landmarks():
    # Create lists of true and predicted landmarks
    true_landmarks = torch.Tensor([[[[1, 2], [3, 4], [5, 6]]]])
    assert true_landmarks.shape == (1, 1, 3, 2)
    pred_landmarks = [torch.tensor([1, 2]), torch.tensor([3, 4]), torch.tensor([7, 8])]

    # Call the function
    pe, tp, fp, fn, _ = multi_instance_point_error(true_landmarks, pred_landmarks)

    # Check that the output has the expected values
    assert pe > 0
    assert tp == 3
    assert fp == 0


def test_multi_instance_point_error_different_landmarks_3d():
    # Create lists of true and predicted landmarks
    true_landmarks = torch.Tensor([[[[1, 2, 4], [3, 4, 5], [5, 6, 3]]]])
    assert true_landmarks.shape == (1, 1, 3, 3)
    pred_landmarks = [torch.tensor([1, 2, 3]), torch.tensor([3, 4, 2]), torch.tensor([7, 8, 5])]

    # Call the function
    pe, tp, fp, fn, _ = multi_instance_point_error(true_landmarks, pred_landmarks)

    # Check that the output has the expected values
    assert pe > 0
    assert tp == 3
    assert fp == 0


def test_multi_instance_point_error_extra_predicted_landmark():
    # Create lists of true and predicted landmarks
    true_landmarks = torch.Tensor([[[[1, 2], [3, 4], [5, 6]]]])
    pred_landmarks = [
        torch.tensor([1, 2]),
        torch.tensor([3, 4]),
        torch.tensor([5, 6]),
        torch.tensor([7, 8]),
    ]

    # Call the function
    pe, tp, fp, fn, _ = multi_instance_point_error(true_landmarks, pred_landmarks)

    # Check that the output has the expected values
    assert pe == 0
    assert tp == 3
    assert fp == 1


def test_multi_instance_point_error_extra_predicted_landmark_3d():
    # Create lists of true and predicted landmarks
    true_landmarks = torch.Tensor([[[[1, 2, 4], [3, 4, 5], [5, 6, 3]]]])
    pred_landmarks = [
        torch.tensor([1, 2, 4]),
        torch.tensor([3, 4, 5]),
        torch.tensor([5, 6, 3]),
        torch.tensor([7, 8, 2]),
    ]

    # Call the function
    pe, tp, fp, fn, _ = multi_instance_point_error(true_landmarks, pred_landmarks)

    # Check that the output has the expected values
    assert pe == 0
    assert tp == 3
    assert fp == 1
